!
! Module: bkgfit_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This task fits a linear combination of background maps to an image.
! 
! Comments:
!	TODO:
!		- try change from working in coeffs to ln(coeffs). Advantages: (i) can't go below zero; (ii) curvature may be positive over a wider area.
!		- trim off blank border during processing, a la asmooth.
!		- Add uncertainty (and/or quasi-chi^2?) calculation.
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module bkgfit_mod

  use dal
  use param  
  use errorHandling

  use dal_aux
  use detection_stats
  use linear
use dump
  use ssc_misc
  use array_utils
  use wcs_aux



!use math_utils



  implicit none 

  private 

  public bkgfit

  ! Command-line parameters:
  character(255) ::&
    inImageSetName,&
!    tempImageSetName,&
    truncImageSetName,&
    maskSetName,&
    outBkgMapSetName
  character(255), allocatable ::&
    fittingImageSetName(:)
  character(20) ::&
    fittingStyle
!  integer ::&
!    numImages
  logical ::&
    withMask,&
    withInitialCoeffs
  real, allocatable ::&
    initialCoeffs(:)!,&
!    userTruncImage(:,:)

  ! Other global variables:
  real(single), parameter :: maxCosDiffAngle = 0.99

contains 

  !---------------------------------------------------------------------
  subroutine bkgfit
  !---------------------------------------------------------------------

    implicit none 

    integer(int16), pointer :: poissImage(:,:)
    integer :: numImages, lengthLongestImageFileName, imageNum, degFree
    integer(int16) :: xSize, ySize
    real(single), allocatable ::&
      imageCube(:,:,:), coeffsOut(:), tempTruncImage(:,:), errorMatrix(:,:)
    real(single), pointer ::&
      tempImage(:,:), arrayDataReal32(:,:), weightsPtr(:), userTruncImage(:,:),&
      errorData(:,:)
!    logical(bool), pointer :: mask(:,:)
!    logical(bool), allocatable :: someImageGtZeroMask(:,:)
    logical, pointer :: mask(:,:)
    logical, allocatable :: someImageGtZeroMask(:,:)
    integer(int32) :: newDimVector(0:1)
    type(DataSetT) :: newSet
    type(ArrayT) :: newArray
    type(TableT) :: tab
    type(ColumnT) :: fileNameCol, errCol
!    type(WcsT) :: wcs
!    logical(bool) :: equivalent
!logical :: doWcsStuff
    real(single) :: chiSquared, chiSquaredProb

    !-------------------------------------------------------------------
    ! Start:

    call readParameters

    numImages = size(fittingImageSetName)
if (numImages <= 0) call error('noComponents',"")

    lengthLongestImageFileName = 0
    do imageNum = 1, numImages
      lengthLongestImageFileName = max(lengthLongestImageFileName&
        , len_trim(fittingImageSetName(imageNum)))
    end do

    call readArrayData(inImageSetName, poissImage)
if (any(poissImage < 0)) call error('negativePoissonValues','')

    if (all(poissImage == 0))&
call error('imageAllZeros',"")

    xSize = size(poissImage, 1)
    ySize = size(poissImage, 2)


    if (withMask) then
      call readArrayData(maskSetName, mask)
      if (size(mask, 1) .ne. xSize .or. size(mask, 2) .ne. ySize)&
call error('badMaskDimensions','')

    else
      allocate(mask(xSize, ySize))
      mask = .true.
    end if

    allocate(coeffsOut(numImages), errorMatrix(numImages,numImages))
    allocate(imageCube(numImages, xSize, ySize))
    allocate(someImageGtZeroMask(xSize, ySize))
    someImageGtZeroMask = .false.
    do imageNum = 1, numImages
      call readArrayData(fittingImageSetName(imageNum), tempImage)
      if (size(tempImage, 1) .ne. xSize .or. size(tempImage, 2) .ne. ySize)&
call error('modelImageWrongSize','')

if (any(tempImage < 0.0)) call error('negativeValuesModelImage',"")
if (all(tempImage == 0.0)) call error('modelImageAllZero',"")

      someImageGtZeroMask = someImageGtZeroMask .or. (tempImage > 0)

      imageCube(imageNum,:,:) = tempImage
      deallocate(tempImage)
    end do

    mask = mask .and. someImageGtZeroMask
    deallocate(someImageGtZeroMask)
if (.not. any(mask)) call error('maskImageAllFalse','')

    ! For each pair of images, check they are not too 'close' (otherwise
    ! the fitting will not work):

    if (.not. imagesHaveEnoughSpread(imageCube, mask, maxCosDiffAngle))&
call error('imagesTooSimilar','')

    ! Now do the fitting:

    select case(fittingStyle)
    case('whole')
      call fitToWholePoisson(imageCube, poissImage, mask, 10&
        , coeffsOut, errorMatrix, chiSquared, degFree, chiSquaredProb)

    case('base')
      call readArrayData(truncImageSetName, userTruncImage)

      if (size(userTruncImage, 1) .ne. xSize&
      .or. size(userTruncImage, 2) .ne. ySize)&
call error('modelImageWrongSize','')

if (any(userTruncImage < 0.0)) call error('negativeValuesTruncationImage','')
if (all(userTruncImage == 0.0)) call error('truncationImageAllZero','')

      call fitToPoissonBase(imageCube, poissImage, userTruncImage, mask, 10&
        , coeffsOut, errorMatrix, chiSquared, degFree, chiSquaredProb)

      deallocate(userTruncImage)

    case('wholethenbase')
      call fitToWholePoisson(imageCube, poissImage, mask, 10&
        , coeffsOut, errorMatrix, chiSquared, degFree, chiSquaredProb)

      allocate(tempTruncImage(xSize, ySize))
      tempTruncImage = 0.0
      do imageNum = 1, numImages
        tempTruncImage = tempTruncImage + coeffsOut(imageNum)&
          * imageCube(imageNum,:,:)
      end do

!call dumpImageToFits(tempTruncImage, 'trunc.ds')

      call fitToPoissonBase(imageCube, poissImage, tempTruncImage, mask, 10&
        , coeffsOut, errorMatrix, chiSquared, degFree, chiSquaredProb)

      deallocate(tempTruncImage)

    case default
call error('badFitStyle','')
    end select

    !...................................................................
    ! Output the results:

!***** why not use saveImage????

    newDimVector = (/xSize, ySize/)

    newSet = dataSet(outBkgMapSetName, CREATE)
    newArray = addArray(newSet, "PRIMARY", REAL32, newDimVector)
!    if (doWcsStuff) call writeWcs(newArray, wcs) ! in ssclib/wcs_aux

    arrayDataReal32 => real32Array2Data(newArray)

    arrayDataReal32 = 0.0
    do imageNum = 1, numImages
      arrayDataReal32  = arrayDataReal32 + imageCube(imageNum,:,:)&
        * coeffsOut(imageNum)
    end do

    call release(newArray)
    deallocate(imageCube)

do imageNum = 1, numImages
print *, 'coeff', imageNum, ':', coeffsOut(imageNum), '+-'&
  , sqrt(errorMatrix(imageNum,imageNum))
end do

    tab = addTable(newSet, 'COMPDATA', int(numImages))
    weightsPtr => real32Data(addColumn(tab, 'WEIGHT', REAL32))
    weightsPtr = coeffsOut
    fileNameCol = addColumn(tab, 'FILENAME', STRING&
      , dimensions=(/int(lengthLongestImageFileName)/))

    do imageNum = 1, numImages
      call setStringCell(fileNameCol, imageNum-1, fittingImageSetName(imageNum))
    end do

    errCol = addColumn(tab, 'VARIANCE', REAL32, '', (/numImages/))
    errorData => real32VectorData(errCol)
    errorData = errorMatrix

    call setAttribute(tab, 'CHISQUAR', chiSquared&
      , comment='Chi squared value.')
    call setAttribute(tab, 'REDCHISQ', chiSquared/real(degFree)&
      , comment='Reduced chi squared value.')
    call setAttribute(tab, 'CHI2PROB', chiSquaredProb&
      , comment='Prob that CHISQUAR value would arise by chance.')
    call setAttribute(tab, 'DEG_FREE', degFree&
      , comment='Chi squared degrees of freedom.')

    call release(newSet)

    deallocate(poissImage, mask)

  end subroutine bkgfit
  
  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none 

    integer :: numImages, imageNum

    !-------------------------------------------------------------------
    ! Start:

    inImageSetName = stringParameter('poissonset')
    numImages = parameterCount('fitsets')
    allocate(fittingImageSetName(numImages))
    do imageNum = 1, numImages
      fittingImageSetName(imageNum) = stringParameter('fitsets', imageNum - 1)
    end do

    fittingStyle = stringParameter('fitstyle')
    select case(fittingStyle)
    case('whole')
    case('base')
      truncImageSetName = stringParameter('truncationset')
    case('wholethenbase')
    case default
call error('badFitStyle',"")
    end select

    withMask = booleanParameter('withmaskset')
    maskSetName = stringParameter('maskset')

    withInitialCoeffs = booleanParameter('withinitcoeffs')
    if (withInitialCoeffs) then
if (parameterCount('initcoeffs') .ne. numImages) call error('wrongNumCoeffs','')
      allocate(initialCoeffs(numImages))
      do imageNum = 1, numImages
        initialCoeffs(imageNum) = realParameter('initcoeffs', imageNum - 1)
      end do
    end if

!    tempImageSetName = stringParameter('tempset')
    outBkgMapSetName = stringParameter('outbkgmapset')

  end subroutine readParameters

  !---------------------------------------------------------------------
  logical function imagesHaveEnoughSpread(imageCube, mask, maxCosDiffAngle)
  !---------------------------------------------------------------------

    implicit none 

    real(single), intent(in) :: imageCube(:,:,:)
!    logical(bool), intent(in) :: mask(size(imageCube,2)&
    logical,      intent(in) :: mask(size(imageCube,2)&
                                    ,size(imageCube,3))
    real(single), intent(in) :: maxCosDiffAngle

    integer :: numImages, imageNum, imageNumA, imageNumB
    real(double) :: imageEuclidNorm(size(imageCube,1)), cosDiffAngle
    real(double) :: tempCube(size(imageCube,1)&
                            ,size(imageCube,2)&
                            ,size(imageCube,3))

    !-------------------------------------------------------------------
    ! Start:

    imagesHaveEnoughSpread = .true. ! outer default

!print *, '        sum:', sum(imageCube(1,:,:), mask)
!print *, 'precise sum:', preciseSum(imageCube(1,:,:), mask)
!print *, '        sum:', sum(real(imageCube(1,:,:), double), mask)
!print *, 'precise sum:', preciseSum(real(imageCube(1,:,:), double), mask)
!call error('','')

    numImages = size(imageCube, 1)

!call dumpImageToFits(imageCube(1,:,:), 'xx_image_1')
!call dumpImageToFits(imageCube(2,:,:), 'xx_image_2')
!call dumpImageToFits(mask, 'xx_mask')
!print *, 'numImages',numImages

    tempCube = imageCube

    if (numImages > 1) then
      do imageNum = 1, numImages
        imageEuclidNorm(imageNum) = sqrt(sum(tempCube(imageNum,:,:)&
          * tempCube(imageNum,:,:), mask))

        tempCube(imageNum,:,:) = tempCube(imageNum,:,:)&
          / imageEuclidNorm(imageNum)
      end do

      do imageNumA = 1, numImages - 1
        do imageNumB = imageNumA + 1, numImages
          cosDiffAngle = sum(tempCube(imageNumA,:,:)&
            * tempCube(imageNumB,:,:), mask)
          ! a.b = |a||b|cos(angle between a and b)

!print *, 'cosDiffAngle, maxCosDiffAngle', cosDiffAngle, maxCosDiffAngle
          if (abs(cosDiffAngle) > maxCosDiffAngle) then
!*** be careful to make maxCosDiffAngle greater than the numeric noise.
            imagesHaveEnoughSpread = .false.
  return
          end if
        end do
      end do
    end if

  end function imagesHaveEnoughSpread

  !---------------------------------------------------------------------
  subroutine fitToWholePoisson(imageCube, poissImage, mask, maxNumIterations&
  , presentCoeffs, errorMatrix, chiSquared, degFree, chiSquaredProb)
  !---------------------------------------------------------------------
  ! This routine fits N background maps (contained in imageCube) to the integer data in poissImage, and returns the N coefficients of the fit. The fit is performed by maximizing a likelihood function which computes, for each image pixel, the probability that the counts in that pixel of poissImage have arisen from a Poisson distribution about an average value given by the coefficient-weighted sum of the background maps at that pixel.

    implicit none 

    real(single),   intent(in)  :: imageCube(:,:,:)
    integer(int16), intent(in)  :: poissImage(size(imageCube,2)&
                                             ,size(imageCube,3))
!    logical(bool),  intent(in)  :: mask(size(imageCube,2)&
    logical,        intent(in)  :: mask(size(imageCube,2)&
                                       ,size(imageCube,3))
    integer,        intent(in)  :: maxNumIterations
    real(single),   intent(out) :: presentCoeffs(size(imageCube,1)),&
                                   errorMatrix(&
                                     size(imageCube,1),&
                                     size(imageCube,1)),&
                                   chiSquared,&
                                   chiSquaredProb
    integer,        intent(out) :: degFree

    integer ::&
      numCoeffs, xSize, ySize, coeffNum, coeffNumA, coeffNumB,&
      iterationNum, foreignStatus, halvingNum
    real(single) ::&
      matrixOf2ndDerivs( size(imageCube,1),size(imageCube,1)),&
!      inverseMatrix(     size(imageCube,1),size(imageCube,1)),&
      gradient(          size(imageCube,1)),&
      coeffDisplacements(size(imageCube,1)),&
      sumImage(     size(imageCube,2),size(imageCube,3)),&
      sumImageSqrd( size(imageCube,2),size(imageCube,3))
    real(single) :: f, oldF, convCriterion

    !-------------------------------------------------------------------
    ! Start:

    convCriterion = sqrt(10.0**(-precision(f)))

    numCoeffs = size(imageCube, 1)
    xSize = size(imageCube,2)
    ySize = size(imageCube,3)

    if (withInitialCoeffs) then
      presentCoeffs = initialCoeffs
    else
      call calcStartingCoeffs(imageCube, poissImage, mask, presentCoeffs)
    end if

    !...................................................................
    ! The following is a pretty standard minimization routine which uses
    ! the formula
    !
    !		grad_i = C_ij (x_j - x0_j)			(1)
    !
    ! to find a minimum in the function f(x_j), where (x_j - x0_j) is the
    ! displacement vector from the position of the minimum, C_ij is the
    ! matrix of second derivatives of f at x_j and grad_i is the gradient
    ! vector of f at x_j. The formula is exact for quadratic f; for other
    ! functional forms the resulting displacement is only approximate, the
    ! approximation improving the nearer to the minimum. The formula is
    ! thus used iteratively to approach the minimum.
    !
    ! In the present case we want to fit a linear combination of real-
    ! valued images B (eg background map components) to a Poissonized
    ! integer image I. The function f to be minimized is constructed as
    !
    !		f(x_i) = -2*ln(P(x_i))
    !
    ! where P is the net probability that the Poisson image arises from
    ! the summed bkg images, given by
    !		         __
    !		P(x_i) = || P_Poisson_jk(x_i).
    !		        j,k
    !
    ! Here P_Poisson_jk is the probability that the value I_jk at the
    ! (j,k)th pixel of the Poissonized image arises from the 'average' value
    ! at that pixel, given by the weighted sum S_jk(x_i) = sum(x_i*B_i_jk)
    ! of the bkg images. In other words,
    !
    !		                     S_jk(x_i)**I_jk * exp(-S_jk[x_i])
    !		P_Poisson_jk(x_i) = -----------------------------------.
    !		                              I_jk!
    ! Hence
    !		            __
    !		f(x_i) = -2 >_ (I_jk * ln[S_jk{x_i}] - S_jk{x_i} - ln[I_jk!]).
    !		            jk
    !
    ! The gradient grad_i is easily calculated as
    !
    !		           __ (         [       I_jk    ])
    !		grad_m = 2 >_ (B_m_jk * [1 - -----------])
    !		           jk (         [     S_jk{x_i} ])
    !
    ! and the second derivatives as
    !
    !		         __ ( B_m_jk * B_n_jk * I_jk )
    !		C_mn = 2 >_ (------------------------).
    !		         jk (      S_jk{x_i}**2      )
    !
    !...................................................................

    call calcFWholePoisson(imageCube, poissImage, mask, presentCoeffs&
      , oldF)

    MAINLOOP: do iterationNum = 1, maxNumIterations
print *, 'iterationNum:', iterationNum

      ! Do some precalculation to save computer time:
      sumImage = 0.0
      do coeffNum = 1, numCoeffs
        sumImage = sumImage + presentCoeffs(coeffNum) * imageCube(coeffNum,:,:)
      end do
      sumImageSqrd = sumImage * sumImage

      ! Now calculate the matrix of derivatives (the curvature matrix) and
      ! the gradient:

      do coeffNumA = 1, numCoeffs
        gradient(coeffNumA) = 2.0 * sum(imageCube(coeffNumA,:,:)&
          - imageCube(coeffNumA,:,:) * poissImage / sumImage, mask)

        do coeffNumB = 1, coeffNumA - 1
          matrixOf2ndDerivs(coeffNumA, coeffNumB)&
            = 2.0 * sum(imageCube(coeffNumA,:,:) * imageCube(coeffNumB,:,:)&
            * poissImage / sumImageSqrd, mask)

          matrixOf2ndDerivs(coeffNumB, coeffNumA)&
            = matrixOf2ndDerivs(coeffNumA, coeffNumB)
        end do
        matrixOf2ndDerivs(coeffNumA, coeffNumA)&
          = 2.0 * sum(imageCube(coeffNumA,:,:) * imageCube(coeffNumA,:,:)&
          * poissImage / sumImageSqrd, mask)
      end do

      ! Invert equation 1:

      call solveLinearEquations(matrixOf2ndDerivs, coeffDisplacements&
        , gradient, pivotPlease=.true., status=foreignStatus)
if (foreignStatus < 0) call error('badLuSolveExitStatus',"")

      halvingNum = 0
      do while(any(presentCoeffs - coeffDisplacements <= 0.0))
        halvingNum = halvingNum + 1
        if (halvingNum > maxNumIterations) then
    exit MAINLOOP
        end if

        coeffDisplacements = coeffDisplacements / 2.0
      end do

      call calcFWholePoisson(imageCube, poissImage, mask&
        , presentCoeffs - coeffDisplacements, f)

      halvingNum = 0
      do while(f > oldF)
        halvingNum = halvingNum + 1
        if (halvingNum > maxNumIterations) then
    exit MAINLOOP
        end if

        coeffDisplacements = coeffDisplacements / 2.0
        call calcFWholePoisson(imageCube, poissImage, mask&
          , presentCoeffs - coeffDisplacements, f)
      end do

      oldF = f

      presentCoeffs = presentCoeffs - coeffDisplacements

      if (sum(coeffDisplacements * coeffDisplacements)&
        / sum(presentCoeffs * presentCoeffs) < convCriterion)&
    exit MAINLOOP

    end do MAINLOOP

!    call invertPosDefMatrix(0.5 * matrixOf2ndDerivs, errorMatrix&
!      , foreignStatus)
!if (foreignStatus < 0) call error('',"")
call invertPosDefMatrix(0.5 * matrixOf2ndDerivs, errorMatrix)	! ssclib/linear

    sumImage = 0.0
    do coeffNum = 1, numCoeffs
      sumImage = sumImage + presentCoeffs(coeffNum) * imageCube(coeffNum,:,:)
    end do
    chiSquared = calcPearsonsChi2(int(poissImage), sumImage, logical(mask))
    degFree = count(mask) - numCoeffs
if (degFree <= 0) call error('',"")

!    call chi2Prob(chiSquared, degFree, chiSquaredProb, foreignStatus)
!call l_chi2Prob(chiSquared, degFree, chiSquaredProb, foreignStatus)
call integratedChi2Prob(chiSquared, degFree, chiSquaredProb, foreignStatus)
    ! in ssclib/detection_stats.

!print *, 'chiSquared, degFree, chiSquaredProb:', chiSquared, degFree, chiSquaredProb

    if (foreignStatus .ne. 0) then
!      print *, 'chi2Prob status return:', foreignStatus
      print *, 'integratedChi2Prob status return:', foreignStatus
call error('badChi2ProbStatus',"")
    end if

  end subroutine fitToWholePoisson

  !---------------------------------------------------------------------
  subroutine calcStartingCoeffs(imageCube, poissImage, inMask, presentCoeffs)
  !---------------------------------------------------------------------

    implicit none 

    real(single),   intent(in)  :: imageCube(:,:,:)
    integer(int16), intent(in)  :: poissImage(size(imageCube,2)&
                                             ,size(imageCube,3))
!    logical(bool), intent(in)  :: inMask(size(imageCube,2)&
    logical,        intent(in)  :: inMask(size(imageCube,2)&
                                         ,size(imageCube,3))
    real(single),   intent(out) :: presentCoeffs(size(imageCube,1))

    real(single), parameter :: countThreshhold = 25.0
!    logical(bool) :: goToPlanB, mask(size(imageCube,2),size(imageCube,3))
    logical :: goToPlanB, mask(size(imageCube,2),size(imageCube,3))
    integer ::&
      numCoeffs, xSize, ySize, coeffNum, binSize,&
      newXSize, newYSize, xi, yi, xxiStart, xxiFinis, yyiStart, yyiFinis,&
      numNonMaskedBins, numBinsAlsoGeThreshhold, status
!    logical(bool), allocatable :: binnedMask(:,:)
    logical, allocatable :: binnedMask(:,:)
    integer(int32), allocatable :: binnedImage(:,:)
    real(single), allocatable :: binnedImageCube(:,:,:)
    real(single) :: poissAve, imageCubeAve, aveCountsPerPixel

    !-------------------------------------------------------------------
    ! Start:

    ! Plan A is to bin up the images so that most of the significant bins have > countThreshhold counts in them, countThreshhold being chosen such that the significant bins have enough counts to make the Gaussian approximation valid. The binned-up bkg map components are then fitted by a standard least-squares procedure.

    goToPlanB = .false. ! default

    numCoeffs = size(imageCube,1)
    xSize = size(imageCube,2)
    ySize = size(imageCube,3)
    mask = .false. ! default
    do coeffNum = 1, numCoeffs
      mask = mask .or. (imageCube(coeffNum,:,:) > 0.0)
    end do
    mask = mask .and. inMask
if (.not. any(mask)) call error('maskImageAllFalse','')

    aveCountsPerPixel = sum(real(poissImage), mask) / real(count(mask))
    binSize = floor(sqrt(2.0 * countThreshhold / aveCountsPerPixel))

    if (binSize > 0) then
      ! Ok this gives us something to work on. But really what we want is a bin size that results in at least 1/2 the non-masked bins (and this 1/2 should also be larger than the number of coeffs) containing counts of countThreshhold or greater. Bin up the image with successive coarseness until either this is achieved or the number of unmasked bins drops below the number of coeffs, or there is only 1 bin in either of the image directions:

      allocate(binnedImage(xSize, ySize), binnedMask(xSize, ySize))
      do ! double binSize each pass:
        newXSize = ceiling(xSize / real(binSize))
        newYSize = ceiling(ySize / real(binSize))

        ! Bin up the mask and the poissImage:
        do xi = 1, newXSize
          xxiStart = 1 + (xi - 1) * binSize
          xxiFinis = min(xi * binSize, xSize)
          do yi = 1, newYSize
            yyiStart = 1 + (yi - 1) * binSize
            yyiFinis = min(yi * binSize, ySize)

            binnedMask(xi,yi)&
              = any(mask(xxiStart:xxiFinis,yyiStart:yyiFinis))
            binnedImage(xi,yi)&
              = sum(poissImage(xxiStart:xxiFinis,yyiStart:yyiFinis)&
              , mask(xxiStart:xxiFinis,yyiStart:yyiFinis))
          end do
        end do

        binnedMask = binnedMask .and. (binnedImage > 0.0) ! because can't
        ! do a least-squares fit to zero-valued image bins, because the
        ! image bin value is used as an estimate of the variance (which
        ! appears in the denominator of the least-squares expression).

        numNonMaskedBins = count(binnedMask(1:newXSize,1:newYSize))
        if (numNonMaskedBins < numCoeffs) then
          goToPlanB = .true.
      exit
        end if

        numBinsAlsoGeThreshhold = count(binnedMask(1:newXSize,1:newYSize)&
          .and. (binnedImage(1:newXSize,1:newYSize) >= countThreshhold))

        if (numBinsAlsoGeThreshhold > numCoeffs&
          .and. numBinsAlsoGeThreshhold > numNonMaskedBins / 2) then
        ! Ok, try the least-squares approach. Bin up the bkg image cube:
          allocate(binnedImageCube(numCoeffs,newXSize,newYSize))
          do xi = 1, newXSize
            xxiStart = 1 + (xi - 1) * binSize
            xxiFinis = min(xi * binSize, xSize)
            do yi = 1, newYSize
              yyiStart = 1 + (yi - 1) * binSize
              yyiFinis = min(yi * binSize, ySize)

              do coeffNum = 1, numCoeffs
                binnedImageCube(coeffNum,xi,yi)&
                  = sum(imageCube(coeffNum,xxiStart:xxiFinis,yyiStart:yyiFinis)&
                  , mask(xxiStart:xxiFinis,yyiStart:yyiFinis))
              end do
            end do
          end do

          call leastSquaresFit(binnedImageCube&
            , real(binnedImage(1:newXSize,1:newYSize), single)&
            , real(binnedImage(1:newXSize,1:newYSize), single)&
            , binnedMask(1:newXSize,1:newYSize)&
            , presentCoeffs, status)

          if (status < 0) goToPlanB = .true.

      exit
        end if

        if (newXSize <= 1 .or. newYSize <= 1) then
          goToPlanB = .true.
      exit
        end if

        binSize = binSize * 2
      end do ! double binSize each pass:

    else ! binSize = 0. Not enough counts.
      goToPlanB = .true.
    end if

    ! Plan B (if there are not enough counts to make plan A worthwhile) is to find the mean value of all bkg images (counting only pixels for which at least one of the bkg components is > 0) and then choose coeff values such that, for N bkg component images, coeff*av for bkg component image i is equal to 1/N of the average of the Poisson image.

    if (goToPlanB) then
      poissAve = sum(poissImage, mask) / real(count(mask))
      do coeffNum = 1, numCoeffs
        imageCubeAve = sum(imageCube(coeffNum,:,:), mask)&
          / real(count(mask))
if (imageCubeAve <= 0.0) call error('negativeValues','')

        presentCoeffs(coeffNum) = poissAve / real(numCoeffs) / imageCubeAve
      end do
    end if

  end subroutine calcStartingCoeffs

  !---------------------------------------------------------------------
  subroutine leastSquaresFit(modelCube, dataImage, varImage, mask&
  , coeffs, status)
  !---------------------------------------------------------------------

    implicit none 

    real(single),     intent(in)  :: modelCube(:,:,:),&
                                     dataImage(size(modelCube,2)&
                                              ,size(modelCube,3)),&
                                     varImage(size(modelCube,2)&
                                             ,size(modelCube,3))
!    logical(bool), intent(in)  :: mask(size(modelCube,2)&
    logical, intent(in)  :: mask(size(modelCube,2)&
                                         ,size(modelCube,3))
    real(single),     intent(out) :: coeffs(size(modelCube,1))
    integer,          intent(out) :: status

    real(single) ::&
      matrix(size(modelCube,1),size(modelCube,1)),&
!      inverseMatrix(size(modelCube,1),size(modelCube,1)),&
      rhsVector(size(modelCube,1))
    integer :: foreignStatus, numCoeffs, coeffNumA, coeffNumB

    !-------------------------------------------------------------------
    ! Start:

    status = 0

    numCoeffs = size(modelCube,1)
    if (count(mask) < numCoeffs) then
      status = -1 ! too few degrees of freedom to constrain variables.
  return
    end if

    if (any(varImage <= 0.0 .and. mask)) then
      status = -2
  return
    end if

    do coeffNumA = 1, numCoeffs
      rhsVector(coeffNumA) = sum(modelCube(coeffNumA,:,:) * dataImage&
        / varImage, mask)

      do coeffNumB = 1, coeffNumA - 1
        matrix(coeffNumA, coeffNumB)&
          = sum(modelCube(coeffNumA,:,:) * modelCube(coeffNumB,:,:)&
          / varImage, mask)

        matrix(coeffNumB, coeffNumA) = matrix(coeffNumA, coeffNumB)
      end do
      matrix(coeffNumA, coeffNumA)&
        = sum(modelCube(coeffNumA,:,:) * modelCube(coeffNumA,:,:)&
        / varImage, mask)
    end do

    call solveLinearEquations(matrix, coeffs&
      , rhsVector, pivotPlease=.true., status=foreignStatus)
    if (foreignStatus < 0) then
      status = foreignStatus - 2
    end if

  end subroutine leastSquaresFit

  !---------------------------------------------------------------------
  subroutine calcFWholePoisson(imageCube, poissImage, mask, presentCoeffs&
  , sumLikelihoods)
  !---------------------------------------------------------------------
  !		           __
  !		f(x_i) = - >_ (I_jk * ln[S_jk{x_i}] - S_jk{x_i} - ln[I_jk!]).
  !		           jk
  !
  !.....................................................................
!*** actually will just calculate it without the ln[I_jk!] for the time being.

    implicit none 

    real(single),     intent(in)  :: imageCube(:,:,:)
    integer(int16),   intent(in)  :: poissImage(size(imageCube,2)&
                                               ,size(imageCube,3))
!    logical(bool), intent(in)  :: mask(size(imageCube,2)&
    logical, intent(in)  :: mask(size(imageCube,2)&
                                         ,size(imageCube,3))
    real(single),     intent(in)  :: presentCoeffs(size(imageCube, 1))
    real(single),     intent(out) :: sumLikelihoods

    integer ::&
      numCoeffs, coeffNum
    real(single) :: sumImage(size(imageCube,2),size(imageCube,3))

    !-------------------------------------------------------------------
    ! Start:

    numCoeffs = size(imageCube, 1)

    sumImage = 0.0
    do coeffNum = 1, numCoeffs
      sumImage = sumImage + presentCoeffs(coeffNum) * imageCube(coeffNum,:,:)
    end do

    sumLikelihoods = -2.0 * sum(poissImage * log(sumImage) - sumImage, mask)

  end subroutine calcFWholePoisson

  !---------------------------------------------------------------------
  subroutine calcGradFWholePoisson(imageCube, poissImage, mask, presentCoeffs&
  , gradient)
  !---------------------------------------------------------------------
  !
  !		         __ (         [       I_jk    ])
  !		grad_m = >_ (B_m_jk * [1 - -----------])
  !		         jk (         [     S_jk{x_i} ])
  !
  !.....................................................................

    implicit none 

    real(single),     intent(in)  :: imageCube(:,:,:)
    integer(int16),   intent(in)  :: poissImage(size(imageCube,2)&
                                               ,size(imageCube,3))
!    logical(bool), intent(in)  :: mask(size(imageCube,2)&
    logical, intent(in)  :: mask(size(imageCube,2)&
                                         ,size(imageCube,3))
    real(single),     intent(in)  :: presentCoeffs(size(imageCube, 1))
    real(single),     intent(out) :: gradient(size(imageCube, 1))

    integer ::&
      numCoeffs, coeffNum
    real(single) :: sumImage(size(imageCube,2),size(imageCube,3))

    !-------------------------------------------------------------------
    ! Start:

    numCoeffs = size(imageCube, 1)

    sumImage = 0.0
    do coeffNum = 1, numCoeffs
      sumImage = sumImage + presentCoeffs(coeffNum) * imageCube(coeffNum,:,:)
    end do

    do coeffNum = 1, numCoeffs
      gradient(coeffNum) = sum(imageCube(coeffNum,:,:)&
        - imageCube(coeffNum,:,:) * poissImage / sumImage, mask)
    end do

    gradient = gradient * 2.0

  end subroutine calcGradFWholePoisson

  !---------------------------------------------------------------------
  subroutine calcCurvatureFWholePoisson(imageCube, poissImage, mask&
  , presentCoeffs, matrixOf2ndDerivs)
  !---------------------------------------------------------------------
  !
  !		       __ ( B_m_jk * B_n_jk * I_jk )
  !		C_mn = >_ (------------------------).
  !		       jk (      S_jk{x_i}**2      )
  !
  !...................................................................

    implicit none 

    real(single),     intent(in)  :: imageCube(:,:,:)
    integer(int16),   intent(in)  :: poissImage(size(imageCube,2)&
                                               ,size(imageCube,3))
!    logical(bool), intent(in)  :: mask(size(imageCube,2)&
    logical, intent(in)  :: mask(size(imageCube,2)&
                                         ,size(imageCube,3))
    real(single),     intent(in)  :: presentCoeffs(size(imageCube, 1))
    real(single),     intent(out) :: matrixOf2ndDerivs(size(imageCube, 1)&
                                                      ,size(imageCube,1))

    integer :: numCoeffs, coeffNum, coeffNumA, coeffNumB
    real(single) ::&
      sumImage(size(imageCube,2),size(imageCube,3)),&
      sumImageSqrd(size(imageCube,2),size(imageCube,3))

    !-------------------------------------------------------------------
    ! Start:

    numCoeffs = size(imageCube, 1)

    sumImage = 0.0
    do coeffNum = 1, numCoeffs
      sumImage = sumImage + presentCoeffs(coeffNum) * imageCube(coeffNum,:,:)
    end do
    sumImageSqrd = sumImage * sumImage

    do coeffNumA = 1, numCoeffs
      do coeffNumB = 1, coeffNumA - 1
        matrixOf2ndDerivs(coeffNumA, coeffNumB)&
          = sum(imageCube(coeffNumA,:,:) * imageCube(coeffNumB,:,:)&
          * poissImage / sumImageSqrd, mask)

        matrixOf2ndDerivs(coeffNumB, coeffNumA)&
          = matrixOf2ndDerivs(coeffNumA, coeffNumB)
      end do
      matrixOf2ndDerivs(coeffNumA, coeffNumA)&
        = sum(imageCube(coeffNumA,:,:) * imageCube(coeffNumA,:,:)&
        * poissImage / sumImageSqrd, mask)
    end do

    matrixOf2ndDerivs = matrixOf2ndDerivs * 2.0

  end subroutine calcCurvatureFWholePoisson

  !---------------------------------------------------------------------
  function calcPearsonsChi2(counts, model, mask) result(chi2)
  !---------------------------------------------------------------------

    implicit none

    integer,      intent(in)           :: counts(:,:)
    real(single), intent(in)           :: model(size(counts,1),size(counts,2))
    logical,      intent(in), optional :: mask(size(counts,1),size(counts,2))
    real(single)                       :: chi2

    integer :: xSize, ySize, xi, yi
    real(single) :: term
    logical, allocatable :: localMask(:,:)

    !-------------------------------------------------------------------
    ! Start

!**** should be in ssclib>3.16.

    chi2 = 0.0
    xSize = size(counts,1)
    ySize = size(counts,2)
    allocate(localMask(xSize,ySize))
    if (present(mask)) then
      localMask = mask
    else
      localMask = .true.
    end if

if (count(localMask) <= 0) call error('maskAllFalse',"")
if (any(counts < 0 .and. localMask)) call error('negativeCounts',"")
if (any(model <= 0 .and. localMask)) call error('nonPositiveModel',"")
  if (all(counts == 0 .and. localMask)) return

    do xi = 1, xSize
      do yi = 1, ySize
      if (.not. localMask(xi,yi)) cycle
        term = counts(xi,yi) - model(xi,yi)
!print *, xi, yi, counts(xi,yi), model(xi,yi), term, chi2
        chi2 = chi2 + term * term / model(xi,yi)
      end do
    end do
!print *, ''

  end function calcPearsonsChi2

  !---------------------------------------------------------------------
  subroutine fitToPoissonBase(imageCube, poissImage, truncationImage, mask&
  , maxNumIterations, coeffs, errorMatrix, chiSquared, degFree, chiSquaredProb)
  !---------------------------------------------------------------------
  ! This is similar to fitToWholePoisson except that pixels with more than a set number of counts are not included in the likelihood calculation. This limit on counts (which may vary over the image) means that probabilities must be calculated with more care than in the fitToWholePoisson routine. Probabilities are still given by the Poisson distribution except that they must be normalized not to 1 but to the sum of probabilities up to the truncation image value.
  !
  ! The thought behind this routine is that it is much easier to conceive of an image arising from the parent background maps with some _added_ component than some _subtracted_ component. Therefore pixels with few counts are more likely to arise solely from the sum of background maps alone than pixels with many counts. Discarding pixels with high counts makes it more likely that the fit will be an accurate one.
  !
  ! Some are must be taken in choosing the truncation image. Clearly a truncation level which is less than 1 is useless - in this case the only possible number of counts in included pixels is zero, and there is no freedom to allow a fit. The truncation level is therefore set at the truncation image value or 1, whichever is the greater.
  !
  ! It can be seen that the algorithm is most effective at separating additions to the background maps when both these additions and the maps generate significantly greater than 1 count per pixel.

    implicit none 

    real(single),   intent(in)  :: imageCube(:,:,:),&
                                   truncationImage(&
                                     size(imageCube,2),size(imageCube,3))
    integer(int16), intent(in)  :: poissImage(&
                                     size(imageCube,2),size(imageCube,3))
!    logical(bool),  intent(in)  :: mask(&
    logical,  intent(in)  :: mask(&
                                     size(imageCube,2),size(imageCube,3))
    integer,        intent(in)  :: maxNumIterations
    real(single),   intent(out) :: coeffs(size(imageCube, 1)),&
                                   errorMatrix(&
                                     size(imageCube,1),&
                                     size(imageCube,1)),&
                                   chiSquared,&
                                   chiSquaredProb
    integer,        intent(out) :: degFree

    integer ::&
      numCoeffs, xSize, ySize
    real(single), allocatable ::&
      presentCoeffs(:),&
      tempTruncImage(:,:)
!    logical(bool), allocatable :: tempMask(:,:)
    logical, allocatable :: tempMask(:,:)

    !-------------------------------------------------------------------
    ! Start:

    numCoeffs = size(imageCube, 1)
    xSize = size(imageCube,2)
    ySize = size(imageCube,3)

    allocate(presentCoeffs(numCoeffs), tempTruncImage(xSize,ySize))
    allocate(tempMask(xSize,ySize))

    if (withInitialCoeffs) then
      presentCoeffs = initialCoeffs
    else
      call calcStartingCoeffs(imageCube, poissImage, mask, presentCoeffs)
    end if

    tempTruncImage = truncationImage
    where(tempTruncImage < 1.0) tempTruncImage = 1.0

    tempMask = mask
    where(poissImage > tempTruncImage) tempMask = .false.

!call dumpImageToFits(tempMask, 'trunc_mask.ds')

    if (.not. imagesHaveEnoughSpread(imageCube, tempMask, maxCosDiffAngle))&
call error('imagesTooSimilar','')

    call minimizePoissonBase(imageCube, poissImage, tempTruncImage&
      , tempMask, maxNumIterations, presentCoeffs, coeffs, errorMatrix&
      , chiSquared, degFree, chiSquaredProb)

  end subroutine fitToPoissonBase

  !---------------------------------------------------------------------
  subroutine minimizePoissonBase(imageCube, poissImage, truncImage&
    , mask, maxNumIterations, oldCoeffs, coeffs, errorMatrix&
    , chiSquared, degFree, chiSquaredProb)
  !---------------------------------------------------------------------

    implicit none 

    real(single),   intent(in)  :: imageCube(:,:,:)
    integer(int16), intent(in)  :: poissImage(size(imageCube,2)&
                                             ,size(imageCube,3))
    real(single),   intent(in)  :: truncImage(size(imageCube,2)&
                                             ,size(imageCube,3))
!    logical(bool),  intent(in)  :: mask(size(imageCube,2)&
    logical,  intent(in)  :: mask(size(imageCube,2)&
                                       ,size(imageCube,3))
    integer(int32), intent(in)  :: maxNumIterations
    real(single),   intent(in)  :: oldCoeffs(size(imageCube,1))
    real(single),   intent(out) :: coeffs(size(imageCube,1)),&
                                   errorMatrix(&
                                     size(imageCube,1),&
                                     size(imageCube,1)),&
                                   chiSquared,&
                                   chiSquaredProb
    integer,        intent(out) :: degFree

    integer :: iterationNum, foreignStatus, halvingNum, coeffNum, numCoeffs
    real(single) ::&
      matrixOf2ndDerivs( size(imageCube,1),size(imageCube,1)),&
      gradient(          size(imageCube,1)),&
      coeffDisplacements(size(imageCube,1)),&
      sumImage(          size(imageCube,2),size(imageCube,3))
    real(single) :: convCriterion, oldF, f

    !-------------------------------------------------------------------
    ! Start:

    numCoeffs = size(imageCube,1)
    convCriterion = sqrt(10.0**(-precision(f)))

    !...................................................................
    ! The following is a pretty standard minimization routine which uses
    ! the formula
    !
    !		grad_i = C_ij (x_j - x0_j)			(1)
    !
    ! to find a minimum in the function f(x_j), where (x_j - x0_j) is the
    ! displacement vector from the position of the minimum, C_ij is the
    ! matrix of second derivatives of f at x_j and grad_i is the gradient
    ! vector of f at x_j. The formula is exact for quadratic f; for other
    ! functional forms the resulting displacement is only approximate, the
    ! approximation improving the nearer to the minimum. The formula is
    ! thus used iteratively to approach the minimum.
    !
    ! In the present case we want to fit a linear combination of real-
    ! valued images B (eg background map components) to a Poissonized
    ! integer image I. The function f to be minimized is constructed as
    !
    !		f(x_i) = -ln(P(x_i))
    !
    ! where P is the net probability that the Poisson image arises from
    ! the summed bkg images, given by
    !		         __
    !		P(x_i) = || P_Poisson_jk(x_i).
    !		        j,k
    !
    ! Here P_Poisson_jk is the probability that the value I_jk at the
    ! (j,k)th pixel of the Poissonized image arises from the 'average' value
    ! at that pixel, given by the weighted sum S_jk(x_i) = sum(x_i*B_i_jk)
    ! of the bkg images; however since only pixels with I less than a
    ! truncation value N_jk are included, the usual Poisson probability must
    ! be normalized to
    !
    !		 __N_jk  S_jk**l * exp(-S_jk)
    !		 >_     ----------------------
    !		 l=0             l!
    !
    ! Under these conditions P_Poisson_jk becomes
    !
    !		                           S_jk(x_i)**I_jk
    !		P_Poisson_jk(x_i) = -------------------------------.
    !		                             __N_jk  S_jk(x_i)**l
    !		                     I_jk! * >_     --------------
    !		                             l=0         l!
    ! Hence
    !
    !		         __ (  [ __N_jk  S_jk{x_i}**l ]                       )
    !		f(x_i) = >_ (ln[ >_     --------------] - I_jk * ln[S_jk{x_i}])+
    !		         jk (  [ l=0         l!       ]                       )
    !
    !		         __
    !		       + >_ ln[I_jk!].
    !		         jk
    !
    ! The gradient is easily calculated as
    !
    !		                     ( __N_jk  S_jk[x_i]**[l-1]                )
    !		                     ( >_     ------------------               )
    !		         __          ( l=1          [l-1]!             I_jk    )
    !		grad_m = >_ B_m_jk * (--------------------------- - -----------)
    !		         jk          (   __N_jk  S_jk[x_i]**l        S_jk[x_i] )
    !		                     (   >_     --------------                 )
    !		                     (   l=0          l!                       )
    !
    ! and the second derivatives as
    !
    !		                            ( __N_jk  S_jk[x_i]**[l-2]
    !		                            ( >_     ------------------
    !		       __                   ( l=2          [l-2]!
    !		C_mn = >_ B_m_jk * B_n_jk * (--------------------------- -
    !		       jk                   (   __N_jk  S_jk[x_i]**l
    !		                            (   >_     --------------
    !		                            (   l=0          l!
    !
    !		        [ __N_jk  S_jk{x_i]**{l-1} ]**2
    !		        [ >_     ------------------]
    !		        [ l=1          {l-1}!      ]
    !		     - --------------------------------- +
    !		          [ __N_jk  S_jk{x_i}**l ]**2
    !		          [ >_     --------------]
    !		          [ l=0          l!      ]
    !
    !		                     )
    !		                     )
    !		            I_jk     )
    !		     + --------------).
    !		        S_jk{x_i}**2 )
    !		                     )
    !		                     )
    !
    ! Note that if N_jk = 1 the first term in C_mn should vanish.
    !...................................................................

    coeffs = oldCoeffs

!print *, '** inside minimizePoissonBase.'
!call dumpImageToFits(logical(mask), 'truncMask_inMPB.ds')

    call calcFPoissonBase(imageCube, poissImage, truncImage, mask&
      , coeffs, oldF)

!print *, 'old x:', coeffs(1)
!print *, 'old f:', oldF

    INNERLOOP: do iterationNum = 1, maxNumIterations
print *, 'iterationNum:', iterationNum
      call calcGradFPoissonBase(imageCube, poissImage, truncImage&
        , mask, coeffs, gradient)

!print *, 'df/dx:', gradient(1)

      call calcCurvatureFPoissonBase(imageCube, poissImage, truncImage&
        , mask, coeffs, matrixOf2ndDerivs)

!print *, 'd2f/dx2:', matrixOf2ndDerivs(1,1)

      ! Invert equation 1:

      call solveLinearEquations(matrixOf2ndDerivs, coeffDisplacements&
        , gradient, pivotPlease=.true., status=foreignStatus)
      if (foreignStatus < 0) then
call error('badLuSolveExitStatus','')
      end if

!print *, 'dx from df/dx / d2f/dx2:', gradient(1) / matrixOf2ndDerivs(1,1)
!print *, '    matrix inversion dx:', coeffDisplacements(1)

      halvingNum = 0
      do while(any(coeffs - coeffDisplacements <= 0.0))
        halvingNum = halvingNum + 1
        if (halvingNum > maxNumIterations) then
          call warning('badDxHalving',"Can't halve displacement dx enough &
            &to reach a non-negative value of x.")
    exit INNERLOOP
        end if

        coeffDisplacements = coeffDisplacements / 2.0
      end do

      call calcFPoissonBase(imageCube, poissImage, truncImage&
        , mask, coeffs - coeffDisplacements, f)

      halvingNum = 0
      do while(f > oldF)
        halvingNum = halvingNum + 1
        if (halvingNum > maxNumIterations) then
          call warning('badDxHalving',"Can't halve displacement dx enough &
            &to reach a lower value of f.")
    exit INNERLOOP
        end if

        coeffDisplacements = coeffDisplacements / 2.0
        call calcFPoissonBase(imageCube, poissImage, truncImage&
          , mask, coeffs - coeffDisplacements, f)
      end do

!print *, '            eventual dx:', coeffDisplacements(1)

      oldF = f

      coeffs = coeffs - coeffDisplacements

!print *, 'x:', coeffs(1)
!print *, 'f:', oldF
!print *, ''

      if (sum(coeffDisplacements * coeffDisplacements)&
        / sum(coeffs * coeffs) < convCriterion) then
print *, '  Inner converged.'
    exit INNERLOOP
      end if

    end do INNERLOOP

!    call invertPosDefMatrix(0.5 * matrixOf2ndDerivs, errorMatrix&
!      , foreignStatus)
!if (foreignStatus < 0) call error('',"")
call invertPosDefMatrix(0.5 * matrixOf2ndDerivs, errorMatrix)	! ssclib/linear

    sumImage = 0.0
    do coeffNum = 1, numCoeffs
      sumImage = sumImage + coeffs(coeffNum) * imageCube(coeffNum,:,:)
    end do
    chiSquared = calcPearsonsChi2(int(poissImage), sumImage, logical(mask))
    degFree = count(mask) - numCoeffs
if (degFree <= 0) call error('',"")

!    call chi2Prob(chiSquared, degFree, chiSquaredProb, foreignStatus)
    call integratedChi2Prob(chiSquared, degFree, chiSquaredProb, foreignStatus)
    ! in ssclib/detection_stats.

    if (foreignStatus .ne. 0) then
!      print *, 'chi2Prob status return:', foreignStatus
      print *, 'integratedChi2Prob status return:', foreignStatus
call error('badChi2ProbStatus',"")
    end if

  end subroutine minimizePoissonBase

  !---------------------------------------------------------------------
  subroutine calcFPoissonBase(imageCube, poissImage, truncImage&
  , mask, presentCoeffs, sumLikelihoods)
  !---------------------------------------------------------------------
  !
  !	         __ (  [ __N_jk  S_jk{x_i}**l ]                       )
  !	f(x_i) = >_ (ln[ >_     --------------] - I_jk * ln[S_jk{x_i}]) +
  !	         jk (  [ l=0         l!       ]                       )
  !
  !	         __
  !	       + >_ ln[I_jk!].
  !	         jk
  !
  !.....................................................................

    implicit none 

    real(single),     intent(in)  :: imageCube(:,:,:)
    integer(int16),   intent(in)  :: poissImage(size(imageCube,2)&
                                               ,size(imageCube,3))
    real(single),     intent(in)  :: truncImage(size(imageCube,2)&
                                               ,size(imageCube,3)),&
                                     presentCoeffs(size(imageCube, 1))
!    logical(bool), intent(in)  :: mask(size(imageCube,2)&
    logical, intent(in)  :: mask(size(imageCube,2)&
                                         ,size(imageCube,3))
    real(single),     intent(out) :: sumLikelihoods

    integer ::&
      numCoeffs, coeffNum, l
    real(double) ::&
      sumImage(size(imageCube,2),size(imageCube,3)),&
      lSumTermImage(size(imageCube,2),size(imageCube,3)),&
      lSumImage(size(imageCube,2),size(imageCube,3)),&
      lnPoissFactImage(size(imageCube,2),size(imageCube,3)),&
!tempImage(size(imageCube,2),size(imageCube,3)),&
      lFactorial

    !-------------------------------------------------------------------
    ! Start:

    numCoeffs = size(imageCube, 1)

    sumImage = 0.0
    do coeffNum = 1, numCoeffs
      sumImage = sumImage + presentCoeffs(coeffNum) * imageCube(coeffNum,:,:)
    end do

    lSumTermImage = 1.0d0
    lFactorial = 1.0d0
    lSumImage = 1.0d0
    do l = 1, floor(maxval(truncImage))
      lSumTermImage = lSumTermImage * sumImage
      lFactorial = lFactorial * l
      where(l <= truncImage)&
        lSumImage = lSumImage + lSumTermImage / lFactorial
    end do

    call calcLnPoissFactImage(poissImage, lnPoissFactImage)
!call dumpImageToFits(lSumImage,        'ju_lSumImage')
!call dumpImageToFits(log(lSumImage),   'ju_log_lSumImage')
!call dumpImageToFits(poissImage,       'ju_poissImage')
!call dumpImageToFits(sumImage,         'ju_sumImage')
!call dumpImageToFits(log(sumImage),    'ju_log_sumImage')
!call dumpImageToFits(lnPoissFactImage, 'ju_lnPoissFactImage')
!call dumpImageToFits(mask,             'ju_mask')

!tempImage = log(lSumImage) - poissImage * log(sumImage)&
!  + lnPoissFactImage

!where(.not. mask) tempImage = 0.0

    sumLikelihoods = 2.0 * sum(log(lSumImage) - poissImage * log(sumImage)&
      + lnPoissFactImage, mask)

!sumLikelihoods = sum(tempImage, mask)

!call dumpImageToFits(tempImage, 'ju_sumLikelihoodsImage')

!print *, 'sumLikelihoods', sumLikelihoods

!call error('','')

  end subroutine calcFPoissonBase

  !---------------------------------------------------------------------
  subroutine calcLnPoissFactImage(poissImage, lnPoissFactImage)
  !---------------------------------------------------------------------

    implicit none 

    integer(int16), intent(in)  :: poissImage(:,:)
    real(double),   intent(out) :: lnPoissFactImage(size(poissImage,1)&
                                                   ,size(poissImage,2))

    integer :: i
    real(double) :: lnIFactorial

    !-------------------------------------------------------------------
    ! Start:

    lnIFactorial = 0.0d0
    lnPoissFactImage = 1.0
    do i = 1, maxval(poissImage)
      lnIFactorial = lnIFactorial + log(real(i))
      where(i == poissImage) lnPoissFactImage = lnIFactorial
    end do

!    lnPoissFactImage = log(lnPoissFactImage)

  end subroutine calcLnPoissFactImage

  !---------------------------------------------------------------------
  subroutine calcGradFPoissonBase(imageCube, poissImage, truncImage&
  , mask, presentCoeffs, gradient)
  !---------------------------------------------------------------------
  !
  !	                     ( __N_jk  S_jk[x_i]**[l-1]                )
  !	                     ( >_     ------------------               )
  !	         __          ( l=1          [l-1]!             I_jk    )
  !	grad_m = >_ B_m_jk * (--------------------------- - -----------).
  !	         jk          (   __N_jk  S_jk[x_i]**l        S_jk[x_i] )
  !	                     (   >_     --------------                 )
  !	                     (   l=0          l!                       )
  !
  !.....................................................................

    implicit none 

    real(single),   intent(in)  :: imageCube(:,:,:)
    integer(int16), intent(in)  :: poissImage(size(imageCube,2)&
                                             ,size(imageCube,3))
    real(single),   intent(in)  :: truncImage(size(imageCube,2)&
                                             ,size(imageCube,3)),&
                                   presentCoeffs(size(imageCube, 1))
!    logical(bool), intent(in)  :: mask(size(imageCube,2)&
    logical, intent(in)  :: mask(size(imageCube,2)&
                                         ,size(imageCube,3))
    real(single),   intent(out) :: gradient(size(imageCube, 1))

    integer ::&
      numCoeffs, coeffNum, l
    real(double) ::&
      sumImage(size(imageCube,2),size(imageCube,3)),&
      lSumTermImage(size(imageCube,2),size(imageCube,3)),&
      lSum0Image(size(imageCube,2),size(imageCube,3)),&
      lSum1Image(size(imageCube,2),size(imageCube,3)),&
      lFactorial

    !-------------------------------------------------------------------
    ! Start:

    numCoeffs = size(imageCube, 1)

    sumImage = 0.0
    do coeffNum = 1, numCoeffs
      sumImage = sumImage + presentCoeffs(coeffNum) * imageCube(coeffNum,:,:)
    end do

    lSumTermImage = 1.0
    lFactorial = 1
    lSum0Image = 1.0
    lSum1Image = 1.0
    do l = 1, floor(maxval(truncImage))
      lSumTermImage = lSumTermImage * sumImage
      lFactorial = lFactorial * l
      where(l <= truncImage)&
        lSum0Image = lSum0Image + lSumTermImage / lFactorial
      where(l <= truncImage - 1.0)&
!        lSum1Image = lSum1Image + lSumTermImage / lFactorial
lSum1Image = lSum0Image
    end do

    do coeffNum = 1, numCoeffs
      gradient(coeffNum) = sum(imageCube(coeffNum,:,:)&
        * (lSum1Image / lSum0Image - poissImage / sumImage), mask)
    end do

    gradient = gradient * 2.0

  end subroutine calcGradFPoissonBase

  !---------------------------------------------------------------------
  subroutine calcCurvatureFPoissonBase(imageCube, poissImage, truncImage&
  , mask, presentCoeffs, matrixOf2ndDerivs)
  !---------------------------------------------------------------------
  !
  !	                            ( __N_jk  S_jk[x_i]**[l-2]
  !	                            ( >_     ------------------
  !	       __                   ( l=2          [l-2]!
  !	C_mn = >_ B_m_jk * B_n_jk * (--------------------------- -
  !	       jk                   (   __N_jk  S_jk[x_i]**l
  !	                            (   >_     --------------
  !	                            (   l=0          l!
  !
  !	        [ __N_jk  S_jk{x_i]**{l-1} ]**2                  )
  !	        [ >_     ------------------]                     )
  !	        [ l=1          {l-1}!      ]            I_jk     )
  !	     - --------------------------------- + --------------).
  !	          [ __N_jk  S_jk{x_i}**l ]**2       S_jk{x_i}**2 )
  !	          [ >_     --------------]                       )
  !	          [ l=0          l!      ]                       )
  !
  ! Note that if N_jk = 1 the first term in C_mn should vanish.
  !...................................................................

    implicit none 

    real(single),   intent(in)  :: imageCube(:,:,:)
    integer(int16), intent(in)  :: poissImage(size(imageCube,2)&
                                             ,size(imageCube,3))
    real(single),   intent(in)  :: truncImage(size(imageCube,2)&
                                             ,size(imageCube,3)),&
                                   presentCoeffs(size(imageCube, 1))
!    logical(bool), intent(in)  :: mask(size(imageCube,2)&
    logical, intent(in)  :: mask(size(imageCube,2)&
                                         ,size(imageCube,3))
    real(single),   intent(out) :: matrixOf2ndDerivs(size(imageCube, 1)&
                                                    ,size(imageCube,1))

    integer :: numCoeffs, coeffNum, coeffNumA, coeffNumB, l
    real(double) ::&
      sumImage(size(imageCube,2),size(imageCube,3)),&
      sumImageSqrd(size(imageCube,2),size(imageCube,3)),&
      lSumTermImage(size(imageCube,2),size(imageCube,3)),&
      lSum0Image(size(imageCube,2),size(imageCube,3)),&
      lSum1Image(size(imageCube,2),size(imageCube,3)),&
      lSum2Image(size(imageCube,2),size(imageCube,3)),&
      lFactorial

    !-------------------------------------------------------------------
    ! Start:

    numCoeffs = size(imageCube, 1)

    sumImage = 0.0
    do coeffNum = 1, numCoeffs
      sumImage = sumImage + presentCoeffs(coeffNum) * imageCube(coeffNum,:,:)
    end do
    sumImageSqrd = sumImage * sumImage

    lSumTermImage = 1.0
    lFactorial = 1
    lSum0Image = 1.0
    lSum1Image = 1.0
    where(0.0 <= truncImage - 2.0)&
      lSum2Image = 1.0
    do l = 1, floor(maxval(truncImage))
      lSumTermImage = lSumTermImage * sumImage
      lFactorial = lFactorial * l
      where(l <= truncImage)&
        lSum0Image = lSum0Image + lSumTermImage / lFactorial
      where(l <= truncImage - 1.0)&
!        lSum1Image = lSum1Image + lSumTermImage / lFactorial
lSum1Image = lSum0Image
      where(l <= truncImage - 2.0)&
!        lSum2Image = lSum2Image + lSumTermImage / lFactorial
lSum2Image = lSum0Image
    end do

    do coeffNumA = 1, numCoeffs
      do coeffNumB = 1, coeffNumA - 1
        matrixOf2ndDerivs(coeffNumA, coeffNumB)&
          = sum(imageCube(coeffNumA,:,:) * imageCube(coeffNumB,:,:)&
          * (lSum2Image / lSum0Image&
          - lSum1Image * lSum1Image / lSum0Image / lSum0Image&
          + poissImage / sumImageSqrd), mask)

        matrixOf2ndDerivs(coeffNumB, coeffNumA)&
          = matrixOf2ndDerivs(coeffNumA, coeffNumB)
      end do
      matrixOf2ndDerivs(coeffNumA, coeffNumA)&
        = sum(imageCube(coeffNumA,:,:) * imageCube(coeffNumA,:,:)&
        * (lSum2Image / lSum0Image&
        - lSum1Image * lSum1Image / lSum0Image / lSum0Image&
        + poissImage / sumImageSqrd), mask)
    end do

    matrixOf2ndDerivs = matrixOf2ndDerivs * 2.0

  end subroutine calcCurvatureFPoissonBase

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template

end module bkgfit_mod
