!+
! Module: eootemap_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This task makes a map of the out-of-time events in an exposure.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module eootemap_mod

  use dal
  use param  
  use errorHandling
  use cal
  use datafilter
  use dss

  use binned_att
  use cube_aux
  use eimchip2sky_aux

  use cal_aux
  use dal_aux
  use intervals_aux	! for gtiVecType
  use save_image
  use ssc_misc
  use wcs_aux
use dump

  implicit none 

  private 

  public eootemap

  character(1024) ::&
    userSelectionExpr=''
  character(256) ::&
!    maskCubeSetName='',&
    expMapCubeSetName='',&
    templateImageSetName='',&
    dssSetName='',&
    ooteCubeSetName='',&
    ooteImageSetName='',&
    srcListSetTabName='',&
    eventSetName='',&
    binnedAttSetTabName = '',&
    tempSetName=''
  character(20) ::&
    style='',&
    outputStyle='',&
    attitudeStyle = '',&
    selectionExprStyle=''
  type(RangeT), allocatable ::&
    userEeVRanges(:)
  logical ::&
    doApprox=.false.
!,&
!    withPileupMask=.false.,&
!    rewriteExpCube=.false.

contains 

  !---------------------------------------------------------------------
  subroutine eootemap
  !---------------------------------------------------------------------

    implicit none 

    type(DataSetT) :: set
    type(ArrayT) :: imArray
    real(single), pointer :: expMapCube(:,:,:), templateImage(:,:)
    type(keyInfoType), pointer :: keyInfo(:)
    ! type def is in eimchip2sky/cube_aux.
    type(gtiVecType), pointer :: gtiVec(:)
    ! type def is in ssclib/intervals_aux.
    integer(int32) :: expmapInstrId, dummyInstrumentId
    type(SpacecraftAttitudeType) ::&
      scAttitude, instrumAttitude, medianScAttitude, dummyInstrumAttitude
    real(double) :: expStartTimeStamp
    integer(int16) :: chipXsize, chipYsize
    integer(int8) :: numCcds, numNodes
    integer :: numCubePlanes, xiSize, yiSize, numAttBins, imageNum
    real(single), allocatable :: ooteCube(:,:,:), binnedAttWeights(:,:), ooteImage(:,:)
    type(WcsT) :: wcs
    character(256) :: binnedAttSetName = ''
    character(20) :: binnedAttTabName = ''
    type(attBinType), pointer :: binnedScAtt(:)
    logical :: noWeightColumn
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start:

    call readParameters

    if (outputStyle .eq. 'sky') then
      !.................................................................
      ! Read some essentials from the template image:

      call message(AppMsg, verbose, "Reading data from template image.")

      set = dataSet(templateImageSetName, READ)
      imArray = array(set, 0)
      call readWcs(set, imArray, wcs) ! in ssclib/wcs_aux.

      call getCalInfo(set, dummyInstrumentId&
        , expStartTimeStamp, medianScAttitude, dummyInstrumAttitude)
      ! in ssclib/cal_aux.

      call readArrayData(imArray, templateImage)
      ! in ssclib/dal_aux.

      call release(set)

      xiSize = size(templateImage, 1)
      yiSize = size(templateImage, 2)
      deallocate(templateImage)

!*** NOTE! the only things that the template image supplies are (i) wcs, (ii) image pixel dimensions, (iii) median attitude and (iv) exposure start time stamp. It would be nice to have some facility to input the last of these in other ways.

      allocate(ooteImage(xiSize,yiSize))
      ooteImage = 0.0
    end if

    !...................................................................

    call readCubeData(expMapCubeSetName, expMapCube, keyInfo, gtiVec)
    ! in eimchip2sky/cube_aux.

    call getCalInfo(expMapCubeSetName, expmapInstrId, expStartTimeStamp&
      , scAttitude, instrumAttitude) ! in ssclib/cal_aux.

!*********** compare expMapCube and templateImage values of instrumentId, expStartTimeStamp, scAttitude??????

    if (expmapInstrId == EPN) then
      chipXSize = EPN_MAX_X	! All these are defined in caloaldefs.
      chipYSize = EPN_MAX_Y
      numCcds = EPN_CCD_LAST
      numNodes = 1

    else if (expmapInstrId == EMOS1 .or. expmapInstrId == EMOS2) then
      chipXSize = EMOS_MAX_X
      chipYSize = EMOS_MAX_Y
      numCcds = EMOS_CCD_LAST
      numNodes = 2

    else
call error('badInstrument',"")
    end if
if (testFlag) print *, 'ccc'

if (minval(keyInfo%ccdNum) < 1) call error('',"")
if (maxval(keyInfo%ccdNum) > numCcds) call error('',"")
if (minval(keyInfo%nodeNum) < 1) call error('',"")
if (maxval(keyInfo%nodeNum) > numNodes) call error('',"")

    numCubePlanes = size(keyInfo)
if (numCubePlanes .ne. size(gtiVec)) call error('wrongNumGtiBlocks',"")
if (numCubePlanes .ne. size(expMapCube, 1)) call error('wrongNumCubeLayers',"")
if (chipXSize     .ne. size(expMapCube, 2)) call error('wrongCubeXSize',"")
if (chipYSize     .ne. size(expMapCube, 3)) call error('wrongCubeYSize',"")

    if (outputStyle .eq. 'sky') then
      !.................................................................
      ! Construct the list of attitudes:

      call message(AppMsg, verbose, "Constructing attitude samples.")
      select case(attitudeStyle)
      case('binnedset')
        call splitSetTabName(binnedAttSetTabName, binnedAttSetName, binnedAttTabName)
        ! in ssclib/dal_aux.
        call readBinnedAttitude(binnedAttSetName, binnedAttTabName, binnedScAtt, noWeightColumn)
        ! in attcalc2/binned_att.
        numAttBins = size(binnedScAtt)

      case('template')
        numAttBins = 1
        allocate(binnedScAtt(1))
        call initializeBinnedAtt(binnedScAtt)
        binnedScAtt(1)%isGood = .true.
        binnedScAtt(1)%gtiDuration = 1.0d0 ! dummy value.
        binnedScAtt(1)%att = medianScAttitude

      case default
call error('badAttitudeStyle',"")
      end select

      allocate(binnedAttWeights(numCubePlanes, numAttBins))

      !.................................................................
      ! Construct the attitude vs. image table of weights:

      call message(AppMsg, verbose, "Calculating attitude weights.")

      do imageNum = 1, numCubePlanes
        call gtiWeightBinnedAttitude(gtiVec(imageNum)%gtis, binnedScAtt)
        ! in attcalc2/binned_att.

        binnedAttWeights(imageNum,:) = binnedScAtt%gtiDuration
      end do ! imageNum

!********** normalize?????
    end if

    !...................................................................

!    if (withPileupMask) then
!      puMaskSet = dataSet(maskCubeSetName, READ)
!      instrumentName = stringAttribute(puMaskSet, "INSTRUME")
!      puMaskInstrId = instId(instrumentName)
!      ! in caloalutils/src/caloaldefs.f90.
!
!      call release(puMaskSet)
!
!if (puMaskInstrId .ne. expmapInstrId) call error('badPileupMaskInstrument',"")
!
!      call readCubeData(maskCubeSetName, puMaskCube, puMaskKeyInfo)
!      ! in eimchip2sky/cube_aux.
!
!      if (numCubePlanes .ne. size(puMaskKeyInfo))&
!call error('badPileupKeySize',"")
!      if (any(keyInfo%ccdNum  .ne. puMaskKeyInfo%ccdNum ))&
!call error('badPileupCcdNums',"")
!      if (any(keyInfo%nodeNum .ne. puMaskKeyInfo%nodeNum))&
!call error('badPileupNodeNums',"")
!if (numCubePlanes .ne. size(puMaskCube, 1)) call error('wrongNumCubeLayers',"")
!if (chipXSize     .ne. size(puMaskCube, 2)) call error('wrongCubeXSize',"")
!if (chipYSize     .ne. size(puMaskCube, 3)) call error('wrongCubeYSize',"")
!
!      do cubePlane = 1, numCubePlanes
!        do rawXi = 1, chipXSize
!          if (any(.not.puMaskCube(cubePlane,rawXi,:)))&
!            expMapCube(cubePlane,rawXi,:) = 0.0
!        end do
!      end do
!
!      if (rewriteExpCube) then
!        expMapSet = dataSet(expMapCubeSetName, MODIFY)
!        expMapArray = array(expMapSet, 0)
!        arrayDataReal32 => real32Array3Data(expMapArray)
!        arrayDataReal32 = expMapCube
!        call release(expMapArray)
!        call release(expMapSet)
!      end if
!
!      deallocate(puMaskCube, puMaskKeyInfo)
!    end if

    ! Now make the OOTE cube:

    allocate(ooteCube(numCubePlanes, chipXSize, chipYSize))

    select case(style)
    case('evlist')
      call makeOoteCubeFromEvlist(keyInfo, expmapInstrId, ooteCube)

    case('srclist')
      call makeOoteCubeFromSrclist(ooteCube)

    case default
call error('badStyle',"")
    end select

    ooteCube = ooteCube * expMapCube

    deallocate(expMapCube)

    !...................................................................
    ! Output:

    select case(outputStyle)
    case('sky')
      call readWcs(templateImageSetName, wcs) ! in ssclib/wcs_aux.

!      call writeCubeToSkyImage(ooteCube&
!        , keyInfo, expStartTimeStamp&
!        , gtiVec, templateImageSetName, templateImageSetName&
!        , ooteImageSetName&
!!        , tempSetName, 'counts', .true., wcs, doApprox, attitudeStyle&
!!        , binnedAttSetName)
!        , tempSetName, 'counts', wcs, doApprox, attitudeStyle&
!        , binnedAttSetName, binnedAttTabName)
!      ! in eimchip2sky/eimchip2sky_aux.

      call rebinChipImagesToSky(ooteCube, keyInfo%ccdNum, wcs, binnedScAtt%att&
        , binnedAttWeights, expStartTimeStamp, 'counts', doApprox, ooteImage)
!rebinChipCubeToSky(chipCube, ccdNums, wcs, binnedScAtt&
!  , binnedAttWeights, expStartTimeStamp, averagingStyle, doApprox, outImage)

      call saveImage(ooteImage, templateImageSetName, ooteImageSetName&
        , tempSetName, wcs=wcs) ! in ssclib/save_image.

      deallocate(binnedScAtt)

    case('raw')
      call writeCubeData(ooteCubeSetName, expMapCubeSetName, ooteCube&
        , keyInfo, gtiVec) ! in eimchip2sky/cube_aux.

    case default
call error('badOutputStyle',"")
    end select

    ! Deallocate all the gtis:
    call deallocateGtiVec(gtiVec) ! in eimchip2sky/cube_aux.

    deallocate(keyInfo)

if (testFlag) print *, 'zzz'

  end subroutine eootemap

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none 

    integer :: numUserRanges, rangeNum

    !-------------------------------------------------------------------
    ! Start:

    expMapCubeSetName = stringParameter('expcubeset')
!    eventSetTabName   = stringParameter('eventtab')

    style = stringParameter('style')
    select case(style)
    case('evlist')
      eventSetName = stringParameter('eventset')
    case('srclist')
      srcListSetTabName = stringParameter('srclisttab')
    case default
call error('badStyle',"")
    end select

    outputStyle = stringParameter('outputstyle')
    select case(outputStyle)
    case('sky')
      templateImageSetName = stringParameter('templateset')
      ooteImageSetName     = stringParameter('ooteimageset')

      attitudeStyle = stringParameter('attstyle')
      select case(attitudeStyle)
      case('binnedset')
        binnedAttSetTabName = stringParameter('binnedatttab')

      case('template')
      case default
call error('badAttitudeStyle',"")
      end select

    case('raw')
      ooteCubeSetName = stringParameter('ootecubeset')
    case default
call error('badOutputStyle',"")
    end select

    selectionExprStyle = stringParameter('selexprstyle')
    select case(selectionExprStyle)
    case('userranges')
      numUserRanges = parameterCount('evlo')
if (numUserRanges .ne. parameterCount('evhi')) call error('wrongNumEvLo',"")
if (numUserRanges <= 0) call error('tooFewUserRanges',"")
      allocate(userEeVRanges(numUserRanges))
      userEeVRanges%lower%type = INCLUSIVE
      userEeVRanges%upper%type = INCLUSIVE
      do rangeNum = 1, numUserRanges
        userEeVRanges(rangeNum)%lower%value = realParameter('evlo', rangeNum-1)
        userEeVRanges(rangeNum)%upper%value = realParameter('evhi', rangeNum-1)
      end do

    case('userexpr')
      userSelectionExpr = stringParameter('expression')

    case('dss')
      dssSetName = stringParameter('dssset')
    case default
call error('badSelExprStyle',"")
    end select

doApprox=.true.
!*****

    tempSetName = stringParameter('tempset')
!tempSetName = 'tempset.ds'

!    withPileupMask = booleanParameter('withpumaskcubeset')
!    if (withPileupMask) then
!      maskCubeSetName = stringParameter('pumaskcubeset')
!      rewriteExpCube = booleanParameter('rewriteexpcube')
!    end if

!    sourceListSetTabName = stringParameter('srclisttab')
!    selectionExpression  = stringParameter('expression')

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine makeOoteCubeFromEvlist(keyInfo, expmapInstrId, ooteCube)
  !---------------------------------------------------------------------
  ! What this should return is a cube with the same number of layers as there are CCDs in the instrument. Each layer is a chip image of the OOTEs for that CCD. Each pixel of such a layer should record the expected OOTE _rate_ at that pixel, _divided by the frame exposure fraction_. The rate is required because at the time the OOTE cube is converted to a sky image (either within the present task or later via a call to eimchip2sky), the OOTE cube layers are multiplied by the total GTIs for that CCD (which are also stored in the cube dataset). The division by frame exposure fraction (=[T_frame-T_readout]/T_frame) is necessary because the OOTE cube must in all cases be multiplied by the exposure-map cube. The values of the exposure map cube are equal to the average frame exposure fraction for the relevant CCD multiplied by a 0 or 1 depending on whether events on that pixel were selected for making the product images or not.
  !
  ! In other words, what we want in the OOTE cube which the task writes to its output is as follows:
  !
  !	            __max_y
  !	            \
  !	            /__   events(c,x,y)
  !	            y=1                       T_readout
  !	o(c,x,y) = ----------------------- * ----------- cts s^-1.
  !	             T_exposure * max_y        T_frame
  !
  ! However since the output p(c,x,y) from the present routine must subsequently be multiplied by f = (T_frame-T_readout)/T_frame, in order for p*f to generate o, clearly we must have
  !
  !
  !	            __max_y
  !	            \
  !	            /__   events(c,x,y)
  !	            y=1                            T_readout
  !	p(c,x,y) = ----------------------- * ---------------------.
  !	             T_exposure * max_y       T_frame - T_readout
  !
  !---------------------------------------------------------------------

    implicit none 

    real(single),      intent(out) :: ooteCube(:,:,:)
    type(keyInfoType), intent(in)  :: keyInfo(size(ooteCube,1))
    integer(int32),    intent(in)  :: expmapInstrId

    integer ::&
      numImages, chipSizeX, chipSizeY, imageNum,&
      ccdNum, nodeNum, ccdNodeNum, rawXi, i, numFilteredEvents
    type(DataSetT) :: eventSet, dssSet
    type(BlockT) :: exposuBlock, dssBlock
    type(DssT) :: dssPointer
    character(1024) :: selectionExpr
    character(20) :: eventInstrName
    integer(int32) :: eventInstrId
    type(TableT) :: eventTab
    integer(int8), pointer :: evlistCcdNum(:)
    integer(int16), pointer :: evlistRawX(:)
    integer(int32), allocatable :: yIntegratedEvents(:,:)
    character(8) :: exposuTabName
    type(ModeParameterDataType) :: modeParams
    character(21) :: piSelStr
    real(double) :: totalGoodTime
character(2) :: ccdStr
logical :: testFlag=.false.

    !-------------------------------------------------------------------
    ! Start:

if (testFlag) print *, '  aaa makeOoteCubeFromEvlist'

    numImages = size(ooteCube, 1)
    chipSizeX = size(ooteCube, 2)
    chipSizeY = size(ooteCube, 3)

if (testFlag) print *, '  bbb'
    ! Filter the event list (same as expmap dss, except for spatial stuff)

    select case(selectionExprStyle)
    case('userranges')
      call createPiSelExprStr(userEeVRanges(1), piSelStr)
      selectionExpr = piSelStr
      do i = 1, size(userEeVRanges)
        call createPiSelExprStr(userEeVRanges(i), piSelStr)
        selectionExpr = trim(selectionExpr)//'||'//piSelStr
      end do

    case('dss')
      dssSet = dataSet(dssSetName, READ)
      dssBlock = block(dssSet, 0)
      dssPointer = dataSubspace(dssBlock)

!!!      selectionExpr = selectExpression(dssPointer, 'PI,PATTERN,CCDNR,TIME')
selectionExpr = selectExpression(dssPointer, 'PI,PATTERN')
      call release(dssSet)

    case('userexpr')
      selectionExpr = userSelectionExpr

    case default
call error('badSelExprStyle',"")
    end select

if (testFlag) print *, '  ccc'

!    call splitSetTabName(eventSetTabName, eventSetName, eventTabName)
!    ! in ssclib/dal_aux.

    eventSet = clone(eventSetName, tempSetName, TEMP)
    eventInstrName = stringAttribute(eventSet, "INSTRUME")
    eventInstrId = instId(eventInstrName) ! in caloalutils/src/caloaldefs.f90
if (eventInstrId .ne. expmapInstrId) call error('inconsistentInstruments',"")

    eventTab = table(eventSet, 'EVENTS', MODIFY)

    call initFilter(selectionExpr)
    numFilteredEvents = doFilter(eventTab)

if (testFlag) print *, '  ddd'

    if (numFilteredEvents <= 0)&
call error('noEventsSelected', "No events passed the DSS filtering expression.")

if (testFlag) print *, '  eee'

    evlistCcdNum => int8Data( column(eventTab, 'CCDNR'))
    evlistRawX   => int16Data(column(eventTab, 'RAWX'))

    allocate(yIntegratedEvents(numImages, chipSizeX))

if (testFlag) print *, '  fff'
    do imageNum = 1, numImages
      ccdNum  = keyInfo(imageNum)%ccdNum
      nodeNum = keyInfo(imageNum)%nodeNum
      ccdNodeNum = 10 * (nodeNum - 1) + ccdNum

if (testFlag) print *, 'ccdNodeNum, imageNum:', ccdNodeNum, imageNum
      do rawXi = 1, chipSizeX
        yIntegratedEvents(imageNum,rawXi)&
          = count((evlistCcdNum == ccdNodeNum).and.(evlistRawX == rawXi))

        ooteCube(imageNum,rawXi,:)&
          = yIntegratedEvents(imageNum,rawXi) / real(chipSizeY)
      end do

      ! Multiply by fraction of time taken up by readout:
      write(exposuTabName, '(a6, i2.2)') 'EXPOSU', ccdNodeNum
if (testFlag) print *, 'exposuTabName:', trim(exposuTabName)
    if (.not. hasBlock(eventSet, exposuTabName)) cycle

      exposuBlock = block(eventSet, exposuTabName)
      call CAL_setState(exposuBlock)
      call CAL_getModeParameters(modeParams)

      totalGoodTime = real64Attribute(exposuBlock, 'LIVETIME')
      ! = sum of GTIs times the 'dead time' (which, at least in MOS
      ! parlance, means the average time lost to cosmic rays).

      ooteCube(imageNum,:,:) = ooteCube(imageNum,:,:)&
        * modeParams%smearedFraction / totalGoodTime

      ! Now, when the OOTE cube is multiplied by the expmap cube+gtis, the
      ! result should give the expected OOTE counts in the image.

if (testFlag) then
  write(ccdStr, '(i2.2)') ccdNodeNum
  call dumpImageToFits(ooteCube(  imageNum,:,:), 'oote_ccd'//ccdStr//'.ds')
end if

      call release(exposuBlock)
    end do

    call release(eventSet)
if (testFlag) print *, '  zzz makeOoteCubeFromEvlist'

  end subroutine makeOoteCubeFromEvlist

  !---------------------------------------------------------------------
  subroutine makeOoteCubeFromEvlist_20040908(expMapCube, keyInfo, expmapInstrId&
  , ooteCube)
  !---------------------------------------------------------------------

    implicit none 

    real(single),      intent(in)  :: expMapCube(:,:,:)
    type(keyInfoType), intent(in)  :: keyInfo(size(expMapCube,1))
    integer(int32),    intent(in)  :: expmapInstrId
    real(single),      intent(out) :: ooteCube(&
                                     size(expMapCube,1),&
                                     size(expMapCube,2),&
                                     size(expMapCube,3)&
                                   )

    integer ::&
      numImages, chipSizeX, chipSizeY, imageNum,&
!, numCcds, numNodes, maxCcdNodeNum, ccdNodeCombo,&
!numCcdNodeCombos
      numFilteredEvents, ccdNum, nodeNum, ccdNodeNum, rawXi, i
!    character(256) :: eventSetName
!    character(20) :: eventTabName
    type(DataSetT) :: dssSet, eventSet
    type(BlockT) :: dssBlock, exposuBlock
    type(DssT) :: dssPointer
    character(1024) :: selectionExpr
    character(20) :: eventInstrName
    integer(int32) :: eventInstrId
    type(TableT) :: eventTab
    integer(int8), pointer :: evlistCcdNum(:)
    integer(int16), pointer :: evlistRawX(:)
    integer(int32), allocatable :: yIntegratedEvents(:,:)
    real(single), allocatable :: yIntegratedExposure(:,:)
    character(8) :: exposuTabName
    type(ModeParameterDataType) :: modeParams
    character(21) :: piSelStr
character(2) :: ccdStr
logical :: testFlag=.false.

    !-------------------------------------------------------------------
    ! Start:

if (testFlag) print *, '  aaa makeOoteCubeFromEvlist'

    numImages = size(expMapCube, 1)
    chipSizeX = size(expMapCube, 2)
    chipSizeY = size(expMapCube, 3)
!    numCcds  = size(ccd2ccdnr,1)
!    numNodes = size(ccd2ccdnr,2)
!    maxCcdNodeNum = 10 * (numNodes - 1) + numCcds
!    numCcdNodeCombos = numNodes * numCcds
!if (maxCcdNodeNum .ne. numImages) call error('wrongNumImages',"")
!if (numCcdNodeCombos .ne. numImages) call error('wrongNumImages',"")

if (testFlag) print *, '  bbb'
    ! Filter the event list (same as expmap dss, except for spatial stuff)

    select case(selectionExprStyle)
    case('userranges')
      call createPiSelExprStr(userEeVRanges(1), piSelStr)
      selectionExpr = piSelStr
      do i = 1, size(userEeVRanges)
        call createPiSelExprStr(userEeVRanges(i), piSelStr)
        selectionExpr = trim(selectionExpr)//'||'//piSelStr
      end do

    case('dss')
      dssSet = dataSet(dssSetName, READ)
      dssBlock = block(dssSet, 0)
      dssPointer = dataSubspace(dssBlock)

!!!      selectionExpr = selectExpression(dssPointer, 'PI,PATTERN,CCDNR,TIME')
selectionExpr = selectExpression(dssPointer, 'PI,PATTERN')
      call release(dssSet)

    case('userexpr')
      selectionExpr = userSelectionExpr

    case default
call error('badSelExprStyle',"")
    end select

if (testFlag) print *, '  ccc'

!    call splitSetTabName(eventSetTabName, eventSetName, eventTabName)
!    ! in ssclib/dal_aux.

    eventSet = clone(eventSetName, tempSetName, TEMP)
    eventInstrName = stringAttribute(eventSet, "INSTRUME")
    eventInstrId = instId(eventInstrName) ! in caloalutils/src/caloaldefs.f90
if (eventInstrId .ne. expmapInstrId) call error('inconsistentInstruments',"")

    eventTab = table(eventSet, 'EVENTS', MODIFY)

    call initFilter(selectionExpr)
    numFilteredEvents = doFilter(eventTab)

if (testFlag) print *, '  ddd'

    if (numFilteredEvents <= 0)&
call error('noEventsSelected', "No events passed the DSS filtering expression.")

    evlistCcdNum => int8Data( column(eventTab, 'CCDNR'))
    evlistRawX   => int16Data(column(eventTab, 'RAWX'))

    allocate(yIntegratedEvents(numImages, chipSizeX))
    allocate(yIntegratedExposure(numImages, chipSizeX))

if (testFlag) print *, '  eee'
    do imageNum = 1, numImages
      ccdNum  = keyInfo(imageNum)%ccdNum
      nodeNum = keyInfo(imageNum)%nodeNum
      ccdNodeNum = 10 * (nodeNum - 1) + ccdNum

if (testFlag) print *, 'ccdNodeNum, imageNum:', ccdNodeNum, imageNum
    do rawXi = 1, chipSizeX
      yIntegratedEvents(imageNum,rawXi)&
          = count((evlistCcdNum == ccdNodeNum).and.(evlistRawX == rawXi))

        yIntegratedExposure(imageNum,rawXi)&
          = sum(expMapCube(imageNum,RawXi,:)) / real(chipSizeY)

        if (yIntegratedExposure(imageNum,rawXi) <= 0.0) then
          ooteCube(imageNum,rawXi,:) = 0.0
        else
          ooteCube(imageNum,rawXi,:)&
            = (yIntegratedEvents(imageNum,rawXi)&
            / yIntegratedExposure(imageNum,rawXi) / real(chipSizeY))&
            * expMapCube(imageNum,rawXi,:)
        end if
      end do

      ! Multiply by fraction of time taken up by readout:
      write(exposuTabName, '(a6, i2.2)') 'EXPOSU', ccdNodeNum
if (testFlag) print *, 'exposuTabName:', trim(exposuTabName)
    if (.not. hasBlock(eventSet, exposuTabName)) cycle

      exposuBlock = block(eventSet, exposuTabName)
      call CAL_setState(exposuBlock)
      call CAL_getModeParameters(modeParams)
      ooteCube(imageNum,:,:) = ooteCube(imageNum,:,:)&
        * modeParams%smearedFraction

if (testFlag) then
  write(ccdStr, '(i2.2)') ccdNodeNum
  call dumpImageToFits(ooteCube(  imageNum,:,:), 'oote_ccd'//ccdStr//'.ds')
  call dumpImageToFits(expMapCube(imageNum,:,:), 'expm_ccd'//ccdStr//'.ds')
end if

      call release(exposuBlock)
    end do

    call release(eventSet)
if (testFlag) print *, '  zzz makeOoteCubeFromEvlist'

  end subroutine makeOoteCubeFromEvlist_20040908

  !---------------------------------------------------------------------
  subroutine makeOoteCubeFromSrclist(ooteCube)
  !---------------------------------------------------------------------

    implicit none 

    real(single), intent(out) :: ooteCube(:,:,:)

    !-------------------------------------------------------------------
    ! Start:

ooteCube=0.0 ! just to shut the compiler up
call error('notYetSupported',"")
!*** stored in temp_code until I can give it some proper attention.

  end subroutine makeOoteCubeFromSrclist

  !---------------------------------------------------------------------
  subroutine createPiSelExprStr(userEeVRange, piSelStr)
  !---------------------------------------------------------------------

    implicit none 

    type(RangeT), intent(in) :: userEeVRange
    character(21), intent(out) :: piSelStr

    character(5) :: loEeVIntStr, hiEeVIntStr

    !-------------------------------------------------------------------
    ! Start:

    piSelStr = ''
    loEeVIntStr = intToStr(nint(userEeVRange%lower%value))
    hiEeVIntStr = intToStr(nint(userEeVRange%upper%value))
    piSelStr = '(PI in ['//trim(loEeVIntStr)//':'//trim(hiEeVIntStr)//'])'

!print *, piSelStr

  end subroutine createPiSelExprStr

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template

end module eootemap_mod

