!+
! Module: eootepileupmask_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This task calculates in rudimentary fashion the pileup probability for each pixel and outputs a mask image indicating those pixels where this probability exceeds a set limit.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module eootepileupmask_mod

  use dal
  use param  
  use errorHandling
  use cal

!  use temp_ssclib

  use binned_att
  use cube_aux
  use eimchip2sky_aux

  use array_utils
  use cal_aux
use coordinate ! only needed for applyAttcalcFudge
  use dal_aux
  use dss_aux
  use save_image
  use ssc_misc
  use wcs_aux

  implicit none 

  private 

  public eootepileupmask

  ! Command-line parameters:
  character(256) ::&
    templateImageSetName='',&
    maskSetName='',&
    binnedAttSetTabName='',&
    tempImageSetName='',&
    eventSetName=''
  character(20) ::&
    attitudeStyle = ''
  real(single) ::&
    pileupMaskProbability=0.0

contains 

  !---------------------------------------------------------------------
  subroutine eootepileupmask
  !---------------------------------------------------------------------

    implicit none 

    real(single), pointer :: templateImage(:,:)
    integer(int16) :: xiSize, yiSize, chipXsize, chipYsize, xi, yi
    type(WcsT) :: wcs
    ! type def is in ssclib/wcs_aux.
    type(DataSetT) :: eventSet
    integer(int32) :: instrumentId
    real(double) :: expStartTimeStamp, startTime, finisTime
    type(SpacecraftAttitudeType) :: medianScAttitude, instrumAttitude
    integer(int8) :: numCcds, numNodes, ccdNum, nodeNum
    integer ::&
      maxNumCubePlanes, numAttBins, numCubePlanes, ccdNodeNum, numEvents,&
      i, attBinNum, startRow, finisRow, row, cubePlaneNum
    type(keyInfoType), allocatable :: rawKeyInfo(:)
    ! type def is in eimchip2sky/cube_aux.
    integer(int8), allocatable :: ccdNums(:)
    integer, allocatable :: cubePlaneNums(:)
    type(attBinType), pointer :: binnedAtts(:)
    ! type def is in attcalc2/binned_att.
    logical :: noWeightColumn
    real(single), allocatable ::&
      binnedAttsNumFrames(:,:), puProbPlane(:,:), puProjectedMask(:,:),&
      outImage(:,:)
    real(single) :: maxPileupProb
    type(BlockT) :: eventBlock, exposuBlock
    character(8) :: exposuTabName
    type(ModeParameterDataType) :: modeParams
    ! type def is in cal/caltypes.
    type(IntervalT), pointer :: gtiPtr(:)
    ! type def is in cal/caltypes.
    type(TableT) :: eventTab
    integer(int8), pointer :: evlistCcdNum(:)
    integer(int16), pointer :: evlistRawX(:), evlistRawY(:)
    real(double), pointer :: evlistTimes(:)
    integer(int8), pointer :: ccdNumValuesInEvlist(:)
    integer(int32), allocatable :: eventCountCube(:,:,:)
    character(256) :: binnedAttSetName
    character(20) :: binnedAttTabName
    real(single), allocatable :: dummyAttWeights(:,:)

    !-------------------------------------------------------------------
    ! Start:

    call readParameters

    !...................................................................
    ! Read some essentials from the template image:

    call message(AppMsg, verbose, "Reading data from template image.")

    call readArrayData(templateImageSetName, templateImage) ! in ssclib/dal_aux.
    xiSize = size(templateImage, 1)
    yiSize = size(templateImage, 2)
    deallocate(templateImage)

    call readWcs(templateImageSetName, wcs) ! in ssclib/wcs_aux.

!**** check instrument name is the same as that of the event list. That is because the template-image header is copied to the new mask image dataset.

    !...................................................................
    ! Set up cal for event list:

    call message(AppMsg, verbose, "Setting up cal from the event list.")

    eventSet = dataSet(eventSetName, READ)

    call getCalInfo(eventSet, instrumentId&
      , expStartTimeStamp, medianScAttitude, instrumAttitude)
    ! in ssclib/cal_aux. This call also sets the state of the cal to the
    ! specified instrument.

    if (instrumentId == EPN) then
      chipXSize = EPN_MAX_X	! All these are defined in caloaldefs.
      chipYSize = EPN_MAX_Y
      numCcds = EPN_CCD_LAST
      numNodes = 1

    else if (instrumentId == EMOS1 .or. instrumentId == EMOS2) then
      chipXSize = EMOS_MAX_X
      chipYSize = EMOS_MAX_Y
      numCcds = EMOS_CCD_LAST
      numNodes = 2

    else
call error('badInstrument',"")
    end if

    maxNumCubePlanes = numCcds * numNodes
    allocate(rawKeyInfo(maxNumCubePlanes), ccdNums(maxNumCubePlanes))
    allocate(cubePlaneNums(maxNumCubePlanes))
    rawKeyInfo%gtiTableName = ''
    ccdNums = 0
    cubePlaneNums = 0 ! default

    !...................................................................
    ! Construct the list of attitudes:

    call message(AppMsg, verbose, "Constructing attitude samples.")

    select case (attitudeStyle)
    case('binnedset')
      call splitSetTabName(binnedAttSetTabName, binnedAttSetName, binnedAttTabName) ! in ssclib/dal_aux.
      call readBinnedAttitude(binnedAttSetName, binnedAttTabName, binnedAtts, noWeightColumn)
      ! in attcalc2/binned_att.
      numAttBins = size(binnedAtts)

    case('events')
      numAttBins = 1
      allocate(binnedAtts(1))
      call initializeBinnedAtt(binnedAtts)
      ! in attcalc2/binned_att.
      binnedAtts(1)%att = medianScAttitude
      binnedAtts(1)%isGood = .true.
      binnedAtts(1)%gtiDuration = 1.0d0 ! dummy value.

    case default
call error('badAttitudeStyle',"")
    end select

    allocate(binnedAttsNumFrames(maxNumCubePlanes, numAttBins))
    allocate(dummyAttWeights(maxNumCubePlanes, 1))
    dummyAttWeights = 1.0

    !...................................................................
    ! Set up CCD-specific stuff from event list:

    call message(AppMsg, verbose, "Reading frame data from EXPOSU blocks.")

    eventBlock = block(eventSet, 'EVENTS')
    numCubePlanes = 0
    do ccdNum = 1, numCcds
      do nodeNum = 1, numNodes
        ccdNodeNum = 10 * (nodeNum - 1) + ccdNum

        write(exposuTabName, '(a6, i2.2)') 'EXPOSU', ccdNodeNum
      if (.not. hasBlock(eventSet, exposuTabName)) cycle

        call message(AppMsg, verbose, "  "//exposuTabName)

        numCubePlanes = numCubePlanes + 1
        cubePlaneNums(ccdNodeNum) = numCubePlanes
        ccdNums(   numCubePlanes) = ccdNodeNum
        rawKeyInfo(numCubePlanes)%ccdNum  = ccdNum
        rawKeyInfo(numCubePlanes)%nodeNum = nodeNum

        exposuBlock = block(eventSet, exposuTabName)
        call CAL_setState(exposuBlock)
        call CAL_getModeParameters(modeParams)
        call release(exposuBlock)

        call readGtisFromDss(eventSet, eventBlock, ccdNodeNum, gtiPtr)
        ! in ssclib/dss_aux.
        call gtiWeightBinnedAttitude(gtiPtr, binnedAtts)
        ! in attcalc2/binned_att.

        ! The call to gtiWeightBinnedAttitude here has a different function to the same call in, for example, eimchip2sky. In the latter, the end result is a real-valued image in sky coordinates which is a superposition of contributions, each of which consists of a projection of the detector CCDs at a separate sample of attitude. In this case one needs to know the weight of each attitude sample, which is given by the length of good time spent at this attitude. The routine gtiWeightBinnedAttitude returns these weight values. In the present case however the output sky image is (conceptually at least) a logical-valued mask, formed by performing a logical OR of the contributing masks. There is no need to know weights to perform this OR (except in the special case that the weight was zero). Thus there is no reason to call gtiWeightBinnedAttitude from this point of view.
        !
        ! On the other hand, the constituent masks at each attitude sample are generated by calculating the amount of pileup during the period at which that attitude sample is applicable. The pileup calculation requires the total GTI during that period - hence we find that gtiWeightBinnedAttitude must be called after all. Note however that in the call to rebinChipImagesToSky, the dummyWeights argument is used to simply set the attitude weights to 1.

        deallocate(gtiPtr)

        binnedAttsNumFrames(numCubePlanes,:) = binnedAtts%gtiDuration&
          * 1000.0d0 / modeParams%frameTime ! %frameTime is in milliseconds.
      end do
    end do
    call release(eventBlock)

    !...................................................................
    ! Open the event list table:

    call message(AppMsg, verbose, "Opening the EVENTS table and doing some checks.")

    eventTab = table(eventSet, 'EVENTS', READ)
    numEvents = numberOfRows(eventTab)
    evlistCcdNum => int8Data(  column(eventTab, 'CCDNR'))
    evlistRawX   => int16Data( column(eventTab, 'RAWX'))
    evlistRawY   => int16Data( column(eventTab, 'RAWY'))
    evlistTimes  => real64Data(column(eventTab, 'TIME'))

    ! Do some checks:

    ccdNumValuesInEvlist => findUniqueValues(evlistCcdNum)
    ! in ssclib/array_utils.

    if (minval(ccdNumValuesInEvlist) < 1&
    .or.maxval(ccdNumValuesInEvlist) > maxNumCubePlanes)&
call error('',"")
    do i = 1, size(ccdNumValuesInEvlist)
      ccdNodeNum = ccdNumValuesInEvlist(i)
      if (cubePlaneNums(ccdNodeNum) <= 0)&
call error('',"")
    end do
    deallocate(ccdNumValuesInEvlist)

    if (minval(evlistRawX) < 1&
    .or.maxval(evlistRawX) > chipXSize)&
call error('',"")

    if (minval(evlistRawY) < 1&
    .or.maxval(evlistRawY) > chipYSize)&
call error('',"")

    ! Now count the events and construct the mask:

    call message(AppMsg, verbose, "Processing the events.")

    allocate(eventCountCube( numCubePlanes,chipXSize,chipYSize))
    allocate(puProbPlane(                  chipXSize,chipYSize))
    allocate(puProjectedMask(numCubePlanes,chipXSize))
    allocate(outImage(xiSize, yiSize))

    outImage = 0.0
    maxPileupProb = 0.0
    do attBinNum = 1, numAttBins
      write(errStr, '(a, i3, a, i3)') 'Attitude bin ', attBinNum, ' out of ', numAttBins
      call message(AppMsg, verbose)

      ! Calculate conservative start and end times for the applicability of each attitude sample. Note that it isn't necessary to consider the affect of GTIs on this period because we assume that the event list contains only events within these GTIs. These crudely-calculated start and stop times will therefore net the same number of events as a more sophisticated calculation would.

      if (attBinNum == 1) then
        startTime = minval(evlistTimes)
      else
        startTime = binnedAtts(attBinNum)%binStartTime
      end if
      startRow = smallestTrueIndex(evlistTimes >= startTime) - 1
      ! smallestTrueIndex is in ssclib/array_utils.

      if (attBinNum == numAttBins) then
        finisTime = maxval(evlistTimes)
      else
        finisTime = binnedAtts(attBinNum+1)%binStartTime
      end if
      finisRow = largestTrueIndex(evlistTimes <= finisTime) - 1
      ! largestTrueIndex is in ssclib/array_utils.

      if (startRow > finisRow) then
        call message(AppMsg, verbose, "No events this attitude bin.")
    cycle
      end if

      eventCountCube = 0
      do row = startRow, finisRow
        errStr = sprintEvery(row, numEvents-1, fraction=0.1&
          , message='  Fraction event list processed:')
        ! in ssclib/ssc_misc.

        if (errStr .ne. '') call message(appMsg, verbose)

      if (evlistTimes(row) < startTime .or. evlistTimes(row) >= finisTime) cycle

        cubePlaneNum = cubePlaneNums(evlistCcdNum(row))
        xi = evlistRawX(row)
        yi = evlistRawY(row)

        eventCountCube(cubePlaneNum, xi, yi)&
          = eventCountCube(cubePlaneNum, xi, yi) + 1
      end do ! loop over event rows

      puProjectedMask = 0.0
      do cubePlaneNum = 1, numCubePlanes
        if (binnedAttsNumFrames(cubePlaneNum,attBinNum) <= 0.0) then
          if (any(eventCountCube(cubePlaneNum,:,:) .ne. 0))&
            call warning('eventsOutOfGoodFrameTime',"")
        else
          puProbPlane = eventCountCube(cubePlaneNum,:,:)&
            / binnedAttsNumFrames(cubePlaneNum, attBinNum)

          ! At this point the array puProbPlane holds not the PU
          ! probability but just the average number of events per
          ! frame for this pixel, CCD and attitude interval. The
          ! probability is calculated at the next step:

          puProbPlane = 1.0 - exp(-puProbPlane)
          maxPileupProb = max(maxPileupProb, maxval(puProbPlane))

          do xi = 1, chipXSize
            if (any(puProbPlane(xi,:) >= pileupMaskProbability))&
              puProjectedMask(cubePlaneNum,xi) = 1.0
            ! The mask here is real-valued because this makes it easier
            ! to rebin it to sky coordinates.
          end do ! loop over raw x
        end if
      end do ! loop over cube planes

      call rebinChipImagesToSky(spread(puProjectedMask, 3, chipYSize)&
        , ccdNums(1:numCubePlanes), wcs, (/binnedAtts(attBinNum)%att/)&
        , dummyAttWeights(1:numCubePlanes,:), expStartTimeStamp, 'rate'&
        , .false., outImage)
      ! in eimchip2sky/eimchip2sky_aux.

      ! How does this call work to make a mask? Ok, we have sent to
      ! rebinChipImagesToSky a cube in which a value of 0 represents a
      ! CCD column for which the pileup criterion has not been exceeded
      ! at any pixel during the present attitude interval, and a value
      ! of 1 represents the contrary. This cube is reprojected to sky
      ! coordinates and added to outImage. At the end of the process
      ! outImage will therefore have 0-valued and non-0-valued pixels.
      ! The 0 values represent those places on the sky where the camera
      ! has never had a piled-up column; >0 values represent the contrary.
      ! These >0 areas of outImage are translated into a logical mask
      ! of the piled-up areas during the final call to saveImage.

    end do ! loop over att bins

    call release(eventSet)
    deallocate(binnedAtts)

    write(errStr, '(a, es11.5)') 'Highest pileup probability found was ', maxPileupProb
    call message(AppMsg, Verbose)

    !...................................................................
    ! Save the mask:

    call message(AppMsg, verbose, "Saving the mask image.")

!    call saveImageLogical((outImage > 0), templateImageSetName, maskSetName&
    call saveImage((outImage > 0), templateImageSetName, maskSetName&
      , tempImageSetName) ! for the moment this is in temp_ssclib.

  end subroutine eootepileupmask

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

    templateImageSetName = stringParameter('templateset')
    tempImageSetName = stringParameter('tempset')

    attitudeStyle = stringParameter('attstyle')
    select case(attitudeStyle)
    case('binnedset')
      binnedAttSetTabName = stringParameter('binnedatttab')
    case('events')
    case default
call error('badAttitudeStyle',"")
    end select

    maskSetName     = stringParameter('pumaskset')
    eventSetName        = stringParameter('eventset')
    pileupMaskProbability = realParameter('puprobcut')

    applyAttcalcFudge = booleanParameter('withboresightfudge')
    ! only needed while boresight psi has wrong sign. IMS 2005-11-17

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template

end module eootepileupmask_mod

