!+
! Module: epileupmask_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This task calculates in rudimentary fashion the pileup probability for each pixel and outputs a mask cube indicating those pixels where this probability exceeds a set limit.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module epileupmask_mod

  use dal
  use param  
  use errorHandling
  use cal

  use cube_aux

  use array_utils
  use ssc_misc

  implicit none 

  private 

  public epileupmask

  ! Command-line parameters:
  character(256) ::&
    maskCubeSetName='',&
    puProbCubeSetName='',&
    eventSetName=''
  real(single) ::&
    pileupMaskProbability = 0.0
  logical ::&
    withPuProbCube=.false.

contains 

  !---------------------------------------------------------------------
  subroutine epileupmask
  !---------------------------------------------------------------------

    implicit none 

    type(DataSetT) :: eventSet
    character(20) :: instrumentName
    integer(int32) ::&
      instrumentId, maxNumCubePlanes, numCubePlanes, ccdNodeNum, cubePlaneNum
    integer :: numEvents, row, maxCcdNodeNum
    integer, allocatable :: cubePlaneNums(:)
    integer(int16) :: chipXsize, chipYsize, xi, yi
    integer(int8) :: numCcds, numNodes, ccdNum, nodeNum
    type(keyInfoType), allocatable :: rawKeyInfo(:)
    ! type def is in eimchip2sky/cube_aux.
!    integer, allocatable :: numFrames(:)
    character(8) :: exposuTabName
    type(BlockT) :: exposuBlock
    type(ModeParameterDataType) :: modeParams
    ! type def is in cal/caltypes.
    real(double) :: totalGoodTime
    integer(int32), allocatable :: eventCountCube(:,:,:)
    real(single), allocatable :: puProbCube(:,:,:), numFrames(:)
    type(TableT) :: eventTab
    integer(int8), pointer :: evlistCcdNum(:)
    integer(int16), pointer :: evlistRawX(:), evlistRawY(:)
    logical, allocatable :: puMaskCube(:,:,:)

    !-------------------------------------------------------------------
    ! Start:

    call readParameters

    call message(appMsg, verbose, 'Reading info from event list header.')

    eventSet = dataSet(eventSetName, READ)

    instrumentName = stringAttribute(eventSet, "INSTRUME")
    instrumentId = instId(instrumentName) ! in caloalutils/src/caloaldefs.f90
    if (instrumentId == EPN) then
      chipXSize = EPN_MAX_X	! All these are defined in caloaldefs.
      chipYSize = EPN_MAX_Y
      numCcds = EPN_CCD_LAST
      numNodes = 1

    else if (instrumentId == EMOS1 .or. instrumentId == EMOS2) then
      chipXSize = EMOS_MAX_X
      chipYSize = EMOS_MAX_Y
      numCcds = EMOS_CCD_LAST
      numNodes = 2

    else
call error('badInstrument',"")
    end if

    maxNumCubePlanes = numCcds * numNodes
    allocate(rawKeyInfo(maxNumCubePlanes))
!    call initializeKeyInfoType(rawKeyInfo) ! in eimchip2sky/cube_aux.
rawKeyInfo%gtiType = 0
rawKeyInfo%gtiTableName = ''
    allocate(numFrames(maxNumCubePlanes))

    call CAL_setState(instrument=instrumentId)

    call message(appMsg, verbose, 'Obtaining frame info from the EXPOSU tables of the event list.')

    maxCcdNodeNum = 10 * (numNodes - 1) + numCcds
    allocate(cubePlaneNums(maxCcdNodeNum))
    cubePlaneNums = 0 ! default

    numCubePlanes = 0
    do ccdNum = 1, numCcds
      do nodeNum = 1, numNodes
        ccdNodeNum = 10 * (nodeNum - 1) + ccdNum

        write(exposuTabName, '(a6, i2.2)') 'EXPOSU', ccdNodeNum
      if (.not. hasBlock(eventSet, exposuTabName)) cycle

        numCubePlanes = numCubePlanes + 1
        rawKeyInfo(numCubePlanes)%ccdNum  = ccdNum
        rawKeyInfo(numCubePlanes)%nodeNum = nodeNum
        cubePlaneNums(ccdNodeNum) = numCubePlanes

        exposuBlock = block(eventSet, exposuTabName)
        call CAL_setState(exposuBlock)
        call CAL_getModeParameters(modeParams)

        totalGoodTime = real64Attribute(exposuBlock, 'LIVETIME')
        ! = sum of GTIs times the 'dead time' (which, at least in MOS
        ! parlance, means the average time lost to cosmic rays).

        numFrames(numCubePlanes) = totalGoodTime * 1000.0d0&
          / modeParams%frameTime ! %frameTime is in milliseconds.

        call release(exposuBlock)
      end do
    end do

    call message(appMsg, verbose, 'Reading data from event list.')

    eventTab = table(eventSet, 'EVENTS', READ)
    numEvents = numberOfRows(eventTab)
    evlistCcdNum => int8Data( column(eventTab, 'CCDNR'))
    evlistRawX   => int16Data(column(eventTab, 'RAWX'))
    evlistRawY   => int16Data(column(eventTab, 'RAWY'))

    ! Do some checks:

    ccdNumValuesInEvlist => findUniqueValues(evlistCcdNum) ! in ssclib/array_utils.
    if (minval(ccdNumValuesInEvlist) < 1&
    .or.maxval(ccdNumValuesInEvlist) > maxCcdNodeNum)&
call error('',"")
    do i = 1, size(ccdNumValuesInEvlist)
      ccdNodeNum = ccdNumValuesInEvlist(i)
      if (cubePlaneNums(ccdNodeNum) <= 0)&
call error('',"")
    end do
    deallocate(ccdNumValuesInEvlist)

    if (minval(evlistRawX) < 1&
    .or.maxval(evlistRawX) > chipXSize)&
call error('',"")

    if (minval(evlistRawY) < 1&
    .or.maxval(evlistRawY) > chipYSize)&
call error('',"")

    ! Now count the events:

    allocate(eventCountCube(numCubePlanes,chipXSize,chipYSize))
    eventCountCube = 0
    do row = 0, numEvents-1
      errStr = sprintEvery(row, numEvents-1, fraction=0.1)
      ! in ssclib/ssc_misc.

      if (errStr .ne. '') call message(appMsg, verbose)

      cubePlaneNum = cubePlaneNums(evlistCcdNum(row))
      xi = evlistRawX(row)
      yi = evlistRawY(row)

      eventCountCube(cubePlaneNum, xi, yi) = eventCountCube(cubePlaneNum, xi, yi) + 1
    end do

    call release(eventSet)

!print *, 'maxval(eventCountCube)=', maxval(eventCountCube)

    allocate(puProbCube(numCubePlanes, chipXSize, chipYSize))
    puProbCube = 0.0
    do cubePlaneNum = 1, numCubePlanes
!print *, 'numFrames(',cubePlaneNum,')=', numFrames(cubePlaneNum)
      if (numFrames(cubePlaneNum) <= 0.0) then
        if (any(eventCountCube(cubePlaneNum,:,:) .ne. 0))&
          call warning('eventsOutOfGoodFrameTime',"")
      else
        puProbCube(cubePlaneNum,:,:) = eventCountCube(cubePlaneNum,:,:)&
          / numFrames(cubePlaneNum)
      end if
    end do

!print *, 'maxval(puProbCube)=', maxval(puProbCube)

    call message(appMsg, verbose, 'Writing mask cube.')

    puProbCube = 1.0 - exp(-puProbCube)

    if (withPuProbCube)&
      call writeCubeData(puProbCubeSetName, eventSetName, puProbCube&
        , rawKeyInfo(:numCubePlanes))
      ! in eimchip2sky/cube_aux.

    allocate(puMaskCube(numCubePlanes, chipXSize, chipYSize))
    puMaskCube = (puProbCube < pileupMaskProbability)

    !...................................................................
    ! Output:

    call writeCubeData(maskCubeSetName, eventSetName, puMaskCube&
      , rawKeyInfo(:numCubePlanes))
    ! in eimchip2sky/cube_aux.

  end subroutine epileupmask

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

    maskCubeSetName     = stringParameter('pumaskcubeset')
    eventSetName        = stringParameter('eventset')
    pileupMaskProbability = realParameter('puprobcut')

    withPuProbCube = booleanParameter('writeprobcube')
    if (withPuProbCube)&
      puProbCubeSetName = stringParameter('probcubeset')

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template

end module epileupmask_mod

