!+
! Module: estrayregion_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This task takes a source list and constructs regions such as to mask out areas around bright sources in all EPIC images, also areas of an EPIC MOS image in which there are RGA-scattered events.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module estrayregion_mod

  use dal
  use param  
  use errorHandling
  use datafilter
  use caloaldefs	! for SpacecraftAttitudeType

  use cal_aux
  use constants
  use coordinate
  use dal_aux		! for splitSetTabName
  use wcs_aux

  implicit none 

  private 

  public estrayregion

  character(1024) ::&
    selectionExpression=''
  character(256) ::&
    eventSetName='',&
    templateImageSetName='',&
    regionSetName='',&
    srcListSetTabName='',&
    tempSetName=''
  real(single) ::&
    radiusArcsec=0.0

contains

  !---------------------------------------------------------------------
  subroutine estrayregion
  !---------------------------------------------------------------------

    implicit none 

    integer, parameter :: m1NumPolyPts = 4, m2NumPolyPts = 4
!    real(single), parameter :: radiusArcsec = 160.0
!    real(single), parameter :: radiusArcsec = 200.0

    integer(int32) :: instrId
    real(double) :: expStartTimeStamp
    type(SpacecraftAttitudeType) ::&
      scAttitude, instrumAttitude
    integer :: numPolyPts, numSources, row, sourceNum
    real(double), allocatable ::&
      poly1X(:), poly1Y(:), poly2X(:), poly2Y(:), srcRa(:), srcDec(:),&
      srcDetX(:), srcDetY(:), detX(:,:), detY(:,:), tanX(:,:), tanY(:,:)
    type(WcsT) :: wcs
    character(256) :: srcListSetName
    character(8) :: srcListTabName
    type(DataSetT) :: srcListSet, regionSet, set
    type(TableT) :: srcListTab, regionTab, tab
    real(single), allocatable :: pixX(:,:), pixY(:,:), pixRadii(:)
    real(double), pointer :: srcRaDeg(:), srcDecDeg(:)
    type(ColumnT) :: shapeCol, xCol, yCol, rCol, compCol
    real(single), pointer :: xData(:,:), yData(:,:), rData(:)
    integer(int16), pointer :: compData(:)

    real(double) ::& ! in units of 0.05 arcsec:
!***** SHOULD define then in tangent units.
      m1Poly1X(m1NumPolyPts) = (/-36000.0, -36000.0, -1500.0, -1500.0/),&
      m1Poly1Y(m1NumPolyPts) = (/  -500.0,    500.0,   500.0,  -500.0/),&
      m1Poly2X(m1NumPolyPts) = (/  1500.0,   1500.0, 36000.0, 36000.0/),&
      m1Poly2Y(m1NumPolyPts) = (/  -500.0,    500.0,   500.0,  -500.0/),&
      m2Poly1X(m2NumPolyPts) = (/  -500.0,    500.0,   500.0,  -500.0/),&
      m2Poly1Y(m2NumPolyPts) = (/-36000.0, -36000.0, -1500.0, -1500.0/),&
      m2Poly2X(m2NumPolyPts) = (/  -500.0,    500.0,   500.0,  -500.0/),&
      m2Poly2Y(m2NumPolyPts) = (/  1500.0,   1500.0, 36000.0, 36000.0/)
!    character(8) :: shapeName = 'POLYGON '
!integer(int32), pointer :: srcId(:)

    !-------------------------------------------------------------------
    ! Start:

    call readParameters

    call getCalInfo(templateImageSetName, instrId, expStartTimeStamp&
      , scAttitude, instrumAttitude) ! in ssclib/cal_aux.

    select case(instrId)
    case(EMOS1)
      numPolyPts = m1NumPolyPts
      allocate(poly1X(numPolyPts), poly1Y(numPolyPts))
      allocate(poly2X(numPolyPts), poly2Y(numPolyPts))
      poly1X = m1Poly1X
      poly2X = m1Poly2X
      poly1Y = m1Poly1Y
      poly2Y = m1Poly2Y

    case(EMOS2)
      numPolyPts = m2NumPolyPts
      allocate(poly1X(numPolyPts), poly1Y(numPolyPts))
      allocate(poly2X(numPolyPts), poly2Y(numPolyPts))
      poly1X = m2Poly1X
      poly2X = m2Poly2X
      poly1Y = m2Poly1Y
      poly2Y = m2Poly2Y

    case(EPN)
      call epstrayregion(radiusArcsec, expStartTimeStamp, scAttitude)
  return

    case default
call error('',"")
    end select

!    call readWcs(templateImageSetName, wcs) ! in ssclib/wcs_aux.
set = dataSet(eventSetName, READ)
tab = table(set, 'EVENTS')
call readWcs(tab, 'X', 'Y', wcs) ! in ssclib/wcs_aux.
call release(set)

    call splitSetTabName(srcListSetTabName, srcListSetName, srcListTabName)
    ! in ssclib/dal_aux.
    srcListSet = clone(srcListSetName, tempSetName, TEMP)
    srcListTab = table(srcListSet, srcListTabName, MODIFY)
    call initFilter(selectionExpression)
    numSources = doFilter(srcListTab)

    if (numSources <= 0) then
      call warning('noSourcesSelected'&
        , "No sources passed the specified filtering expression.")

    else if (numSources == 1) then
      call message(AppMsg, Verbose, '1 source passed the &
        &selection expression.')

    else
      write (errStr, '(i6, a)') numSources, ' sources passed the &
        &selection expression.'
      call message(AppMsg, Verbose)
    end if

    allocate(srcRa(  numSources), srcDec( numSources))
    allocate(srcDetX(numSources), srcDetY(numSources))
    allocate(detX(numPolyPts,numSources), detY(numPolyPts,numSources))
    allocate(tanX(numPolyPts,numSources), tanY(numPolyPts,numSources))

    allocate(pixX(numPolyPts,3*numSources), pixY(numPolyPts,3*numSources))
    allocate(pixRadii(3*numSources))

!srcId  => int32Data(column(srcListTab, 'BOX_ID_SRC'))
!do sourceNum = 1, numSources
!  print *, 'sourceNum, box_id:', sourceNum, srcId(sourceNum-1)
!end do

    srcRaDeg  => real64Data(column(srcListTab, 'RA'))
    srcDecDeg => real64Data(column(srcListTab, 'DEC'))
    srcRa  = srcRaDeg  * PiOn180
    srcDec = srcDecDeg * PiOn180
    call release(srcListSet)

    ! Coordinate transforms:

    call raDecToDet(srcRa, srcDec, scAttitude, expStartTimeStamp&
      , srcDetX, srcDetY) ! in ssclib/coordinate.

    pixX     = 0.0 ! This is necessary because only the 1st member
    pixY     = 0.0 ! of X and Y gets a value when SHAPE=CIRCLE, whereas
    pixRadii = 0.0 ! R is only set when SHAPE=CIRCLE.

pixRadii(2*numSources+1:) = radiusArcsec / detUnitArcsec
!*** dodgy - it depends on the DETXY units being the same as the XY. Some day, fix it up so it's nice.

    ! Transform 1st polynomial to image pixel coordinates:

    do sourceNum = 1, numSources
      detX(:,sourceNum) = srcDetX(sourceNum) + poly1X
      detY(:,sourceNum) = srcDetY(sourceNum) + poly1Y
    end do

    call detToTan(detX, detY, wcs%x%refWorld*PiOn180, wcs%y%refWorld*PiOn180&
      , scAttitude, expStartTimeStamp, tanX, tanY)
    ! in ssclib/coordinate.

    call tanToPixels(wcs, tanX, tanY, pixX(:,1:numSources)&
      , pixY(:,1:numSources))
    ! in ssclib/wcs_aux.

    ! Transform 2nd polynomial to image pixel coordinates:

    do sourceNum = 1, numSources
      detX(:,sourceNum) = srcDetX(sourceNum) + poly2X
      detY(:,sourceNum) = srcDetY(sourceNum) + poly2Y
    end do

    call detToTan(detX, detY, wcs%x%refWorld*PiOn180, wcs%y%refWorld*PiOn180&
      , scAttitude, expStartTimeStamp, tanX, tanY)
    ! in ssclib/coordinate.

    call tanToPixels(wcs, tanX, tanY, pixX(:,numSources+1:2*numSources)&
      , pixY(:,numSources+1:2*numSources))
    ! in ssclib/wcs_aux.

    ! Transform circle centres to image pixel coordinates:

    call detToTan(srcDetX, srcDetY, wcs%x%refWorld*PiOn180&
      , wcs%y%refWorld*PiOn180, scAttitude, expStartTimeStamp&
      , tanX(1,:), tanY(1,:))
    ! in ssclib/coordinate.

    call tanToPixels(wcs, tanX(1,:), tanY(1,:)&
      , pixX(1,2*numSources+1:3*numSources)&
      , pixY(1,2*numSources+1:3*numSources))
    ! in ssclib/wcs_aux.


    ! Save the output regions:

!***** can we call region/ims_ascregion_mod.f90:writeRegionExtension instead??

    regionSet = dataSet(regionSetName, CREATE)
    regionTab = addtable(regionSet, 'REGION', 3*numSources)

    call setAttribute(regionTab, "HDUVERS", "1.0.0")
    call setAttribute(regionTab, "HDUCLASS", "ASC")
    call setAttribute(regionTab, "HDUCLAS1", "REGION")
    call setAttribute(regionTab, "HDUCLAS2", "STANDARD")
    call setAttribute(regionTab, "MTYPE1", "pos")
    call setAttribute(regionTab, "MFORM1", "X,Y")

    shapeCol = addColumn(regionTab, 'SHAPE', STRING, dimensions=(/8/))
    xCol     = addColumn(regionTab, 'X',     REAL32, dimensions=(/numPolyPts/))
    yCol     = addColumn(regionTab, 'Y',     REAL32, dimensions=(/numPolyPts/))
    rCol     = addColumn(regionTab, 'R',     REAL32)
    compCol  = addColumn(regionTab, 'COMPONENT', INTEGER16)

!*** call writeWcs(xCol, yCol, wcs) ! in ssclib/wcs_aux.
!*** need this? check with regionmask

    do row = 0, 2*numSources-1
!      call setStringCell(shapeCol, row, shapeName)
      call setStringCell(shapeCol, row, 'POLYGON ')
    end do
    do row = 2*numSources, 3*numSources-1
!      call setStringCell(shapeCol, row, shapeName)
      call setStringCell(shapeCol, row, 'CIRCLE  ')
    end do

    xData => real32VectorData(xCol)
    yData => real32VectorData(yCol)
    rData => real32Data(rCol)
    xData = pixX
    yData = pixY
    rData = pixRadii

    compData => int16Data(compCol)
    compData = (/(row, row=1,3*numSources)/)

    call release(regionSet)

  end subroutine estrayregion

  !---------------------------------------------------------------------
  subroutine epstrayregion(radiusArcsec, expStartTimeStamp, scAttitude)
  !---------------------------------------------------------------------

    implicit none 

    real(single),                 intent(in) :: radiusArcsec
    real(double),                 intent(in) :: expStartTimeStamp
    type(SpacecraftAttitudeType), intent(in) :: scAttitude

    integer :: numSources, row
!, sourceNum
    real(double), allocatable ::&
      srcRa(:), srcDec(:),&
      srcDetX(:), srcDetY(:), tanX(:), tanY(:)
    type(WcsT) :: wcs
    character(256) :: srcListSetName
    character(8) :: srcListTabName
    type(DataSetT) :: srcListSet, regionSet, set
    type(TableT) :: srcListTab, regionTab, tab
    real(single), allocatable :: pixX(:), pixY(:), pixRadii(:)
    real(double), pointer :: srcRaDeg(:), srcDecDeg(:)
    type(ColumnT) :: shapeCol, xCol, yCol, rCol, compCol
    real(single), pointer :: xData(:), yData(:), rData(:)
    integer(int16), pointer :: compData(:)

!integer(int32), pointer :: srcId(:)

    !-------------------------------------------------------------------
    ! Start:


!    call readWcs(templateImageSetName, wcs) ! in ssclib/wcs_aux.
set = dataSet(eventSetName, READ)
tab = table(set, 'EVENTS')
call readWcs(tab, 'X', 'Y', wcs) ! in ssclib/wcs_aux.
call release(set)

    call splitSetTabName(srcListSetTabName, srcListSetName, srcListTabName)
    ! in ssclib/dal_aux.
    srcListSet = clone(srcListSetName, tempSetName, TEMP)
    srcListTab = table(srcListSet, srcListTabName, MODIFY)
    call initFilter(selectionExpression)
    numSources = doFilter(srcListTab)

    if (numSources <= 0) then
      call warning('noSourcesSelected'&
        , "No sources passed the specified filtering expression.")

    else if (numSources == 1) then
      call message(AppMsg, Verbose, '1 source passed the &
        &selection expression.')

    else
      write (errStr, '(i6, a)') numSources, ' sources passed the &
        &selection expression.'
      call message(AppMsg, Verbose)
    end if

    allocate(srcRa(  numSources), srcDec( numSources))
    allocate(srcDetX(numSources), srcDetY(numSources))
    allocate(tanX(   numSources), tanY(   numSources))
    allocate(pixX(   numSources), pixY(   numSources))
    allocate(pixRadii(numSources))

!srcId  => int32Data(column(srcListTab, 'BOX_ID_SRC'))
!do sourceNum = 1, numSources
!  print *, 'sourceNum, box_id:', sourceNum, srcId(sourceNum-1)
!end do

    srcRaDeg  => real64Data(column(srcListTab, 'RA'))
    srcDecDeg => real64Data(column(srcListTab, 'DEC'))
    srcRa  = srcRaDeg  * PiOn180
    srcDec = srcDecDeg * PiOn180
    call release(srcListSet)

    ! Coordinate transforms:

    call raDecToDet(srcRa, srcDec, scAttitude, expStartTimeStamp&
      , srcDetX, srcDetY) ! in ssclib/coordinate.

pixRadii = radiusArcsec / detUnitArcsec
!*** dodgy - it depends on the DETXY units being the same as the XY. Some day, fix it up so it's nice.

    ! Transform circle centres to image pixel coordinates:

    call detToTan(srcDetX, srcDetY, wcs%x%refWorld*PiOn180&
      , wcs%y%refWorld*PiOn180, scAttitude, expStartTimeStamp&
      , tanX, tanY)
    ! in ssclib/coordinate.

    call tanToPixels(wcs, tanX, tanY, pixX, pixY)
    ! in ssclib/wcs_aux.


    ! Save the output regions:

!***** can we call region/ims_ascregion_mod.f90:writeRegionExtension instead??

    regionSet = dataSet(regionSetName, CREATE)
    regionTab = addtable(regionSet, 'REGION', numSources)

    call setAttribute(regionTab, "HDUVERS", "1.0.0")
    call setAttribute(regionTab, "HDUCLASS", "ASC")
    call setAttribute(regionTab, "HDUCLAS1", "REGION")
    call setAttribute(regionTab, "HDUCLAS2", "STANDARD")
    call setAttribute(regionTab, "MTYPE1", "pos")
    call setAttribute(regionTab, "MFORM1", "X,Y")

    shapeCol = addColumn(regionTab, 'SHAPE', STRING, dimensions=(/8/))
    xCol     = addColumn(regionTab, 'X',     REAL32)
    yCol     = addColumn(regionTab, 'Y',     REAL32)
    rCol     = addColumn(regionTab, 'R',     REAL32)
    compCol  = addColumn(regionTab, 'COMPONENT', INTEGER16)

!*** call writeWcs(xCol, yCol, wcs) ! in ssclib/wcs_aux.
!*** need this? check with regionmask

    do row = 0, numSources-1
      call setStringCell(shapeCol, row, 'CIRCLE  ')
    end do

    xData => real32Data(xCol)
    yData => real32Data(yCol)
    rData => real32Data(rCol)
    xData = pixX
    yData = pixY
    rData = pixRadii

    compData => int16Data(compCol)
    compData = (/(row, row=1,numSources)/)

    call release(regionSet)

  end subroutine epstrayregion

  !---------------------------------------------------------------------
  subroutine estrayregion_20041117
  !---------------------------------------------------------------------

!*** this embodies a temporary workaround until SOC SPR 2560 on regionmask is fixed.

    implicit none 

    integer(int32) :: instrId
    real(double) :: expStartTimeStamp
    type(SpacecraftAttitudeType) ::&
      scAttitude, instrumAttitude
    integer :: numPolyPts, numSources, row, sourceNum
    real(double), allocatable ::&
      poly1X(:), poly1Y(:), srcRa(:), srcDec(:),&
      srcDetX(:), srcDetY(:), detX(:,:), detY(:,:), tanX(:,:), tanY(:,:)
    type(WcsT) :: wcs
    character(256) :: srcListSetName
    character(8) :: srcListTabName
    type(DataSetT) :: srcListSet, regionSet
    type(TableT) :: srcListTab, regionTab
    real(single), allocatable :: pixX(:,:), pixY(:,:)
    real(double), pointer :: srcRaDeg(:), srcDecDeg(:)
    type(ColumnT) :: shapeCol, xCol, yCol
    real(single), pointer :: xData(:,:), yData(:,:)

    integer, parameter :: m1NumPolyPts = 12, m2NumPolyPts = 12
    real(double) ::&
      m1Poly1X(m1NumPolyPts) = (/&
        -36000.0, -36000.0, -1500.0, -1500.0, -36000.0, -36000.0,&
         36000.0,  36000.0,  1500.0,  1500.0,  36000.0,  36000.0/),&
      m1Poly1Y(m1NumPolyPts) = (/&
        -36000.0,   -500.0,  -500.0,   500.0,    500.0,  36000.0,&
         36000.0,    500.0,   500.0,  -500.0,   -500.0, -36000.0/),&
      m2Poly1X(m2NumPolyPts) = (/&
        -36000.0, -36000.0, -1500.0, -1500.0, -36000.0, -36000.0,&
         36000.0,  36000.0,  1500.0,  1500.0,  36000.0,  36000.0/),&
      m2Poly1Y(m2NumPolyPts) = (/&
        -36000.0,   -500.0,  -500.0,   500.0,    500.0,  36000.0,&
         36000.0,    500.0,   500.0,  -500.0,   -500.0, -36000.0/)
    character(8) :: shapeName = 'POLYGON '

    !-------------------------------------------------------------------
    ! Start:

    call readParameters

    call getCalInfo(templateImageSetName, instrId, expStartTimeStamp&
      , scAttitude, instrumAttitude) ! in ssclib/cal_aux.

    select case(instrId)
    case(EMOS1)
      numPolyPts = m1NumPolyPts
      allocate(poly1X(numPolyPts), poly1Y(numPolyPts))
      poly1X = m1Poly1X
      poly1Y = m1Poly1Y

    case(EMOS2)
      numPolyPts = m2NumPolyPts
      allocate(poly1X(numPolyPts), poly1Y(numPolyPts))
      poly1X = m2Poly1X
      poly1Y = m2Poly1Y

    case default
call error('',"")
    end select

    call readWcs(templateImageSetName, wcs) ! in ssclib/wcs_aux.

    call splitSetTabName(srcListSetTabName, srcListSetName, srcListTabName)
    ! in ssclib/dal_aux.
    srcListSet = clone(srcListSetName, tempSetName, TEMP)
    srcListTab = table(srcListSet, srcListTabName, MODIFY)
    call initFilter(selectionExpression)
    numSources = doFilter(srcListTab)
    if (numSources <= 0)&
!call error('noSourcesSelected'&
!, "No sources passed the specified filtering expression.")
call warning('noSourcesSelected'&
, "No sources passed the specified filtering expression.")

    allocate(srcRa(  numSources), srcDec( numSources))
    allocate(srcDetX(numSources), srcDetY(numSources))
    allocate(detX(numPolyPts,numSources), detY(numPolyPts,numSources))
    allocate(tanX(numPolyPts,numSources), tanY(numPolyPts,numSources))
    allocate(pixX(numPolyPts,numSources), pixY(numPolyPts,numSources))

    srcRaDeg  => real64Data(column(srcListTab, 'RA'))
    srcDecDeg => real64Data(column(srcListTab, 'DEC'))
    srcRa  = srcRaDeg  * PiOn180
    srcDec = srcDecDeg * PiOn180
    call release(srcListSet)

    ! Coordinate transforms:

    call raDecToDet(srcRa, srcDec, scAttitude, expStartTimeStamp&
      , srcDetX, srcDetY) ! in ssclib/coordinate.

    row = 1
    do sourceNum = 1, numSources
      detX(:,row) = srcDetX(sourceNum) + poly1X
      detY(:,row) = srcDetY(sourceNum) + poly1Y
      row = row + 1
    end do

    call detToTan(detX, detY, wcs%x%refWorld*PiOn180, wcs%y%refWorld*PiOn180&
      , scAttitude, expStartTimeStamp, tanX, tanY) ! in ssclib/coordinate.

    call tanToPixels(wcs, tanX, tanY, pixX, pixY) ! in ssclib/wcs_aux.

    ! Save the output regions:

!***** can we call region/ims_ascregion_mod.f90:writeRegionExtension instead??

    regionSet = dataSet(regionSetName, CREATE)
    regionTab = addtable(regionSet, 'REGION', numSources)

    call setAttribute(regionTab, "HDUVERS", "1.0.0")
    call setAttribute(regionTab, "HDUCLASS", "ASC")
    call setAttribute(regionTab, "HDUCLAS1", "REGION")
    call setAttribute(regionTab, "HDUCLAS2", "STANDARD")
    call setAttribute(regionTab, "MTYPE1", "pos")
    call setAttribute(regionTab, "MFORM1", "X,Y")

    shapeCol = addColumn(regionTab, 'SHAPE', STRING, dimensions=(/8/))
    xCol     = addColumn(regionTab, 'X',     REAL32, dimensions=(/numPolyPts/))
    yCol     = addColumn(regionTab, 'Y',     REAL32, dimensions=(/numPolyPts/))

!*** call writeWcs(xCol, yCol, wcs) ! in ssclib/wcs_aux.
!*** need this? check with regionmask

    do row = 0, numSources*2-1
      call setStringCell(shapeCol, row, shapeName)
    end do

    xData => real32VectorData(xCol)
    yData => real32VectorData(yCol)
    xData = pixX
    yData = pixY

    call release(regionSet)

  end subroutine estrayregion_20041117

  !---------------------------------------------------------------------
  subroutine estrayregion_future
  !---------------------------------------------------------------------

    implicit none 

    integer(int32) :: instrId
    real(double) :: expStartTimeStamp
    type(SpacecraftAttitudeType) ::&
      scAttitude, instrumAttitude
    integer :: numPolyPts, numSources, row, sourceNum
    real(double), allocatable ::&
      poly1X(:), poly1Y(:), poly2X(:), poly2Y(:), srcRa(:), srcDec(:),&
      srcDetX(:), srcDetY(:), detX(:,:), detY(:,:), tanX(:,:), tanY(:,:)
    type(WcsT) :: wcs
    character(256) :: srcListSetName
    character(8) :: srcListTabName
    type(DataSetT) :: srcListSet, regionSet
    type(TableT) :: srcListTab, regionTab
    real(single), allocatable :: pixX(:,:), pixY(:,:)
    real(double), pointer :: srcRaDeg(:), srcDecDeg(:)
    type(ColumnT) :: shapeCol, xCol, yCol
    real(single), pointer :: xData(:,:), yData(:,:)

    integer, parameter :: m1NumPolyPts = 4, m2NumPolyPts = 4
    real(double) ::&
      m1Poly1X(m1NumPolyPts) = (/-36000.0, -36000.0, -1500.0, -1500.0/),&
      m1Poly1Y(m1NumPolyPts) = (/  -500.0,    500.0,   500.0,  -500.0/),&
      m1Poly2X(m1NumPolyPts) = (/  1500.0,   1500.0, 36000.0, 36000.0/),&
      m1Poly2Y(m1NumPolyPts) = (/  -500.0,    500.0,   500.0,  -500.0/),&
      m2Poly1X(m2NumPolyPts) = (/  -500.0,    500.0,   500.0,  -500.0/),&
      m2Poly1Y(m2NumPolyPts) = (/-36000.0, -36000.0, -1500.0, -1500.0/),&
      m2Poly2X(m2NumPolyPts) = (/  -500.0,    500.0,   500.0,  -500.0/),&
      m2Poly2Y(m2NumPolyPts) = (/  1500.0,   1500.0, 36000.0, 36000.0/)
    character(8) :: shapeName = '!POLYGON'

    !-------------------------------------------------------------------
    ! Start:

    call readParameters

    call getCalInfo(templateImageSetName, instrId, expStartTimeStamp&
      , scAttitude, instrumAttitude) ! in ssclib/cal_aux.

    select case(instrId)
    case(EMOS1)
      numPolyPts = m1NumPolyPts
      allocate(poly1X(numPolyPts), poly1Y(numPolyPts))
      allocate(poly2X(numPolyPts), poly2Y(numPolyPts))
      poly1X = m1Poly1X
      poly2X = m1Poly2X
      poly1Y = m1Poly1Y
      poly2Y = m1Poly2Y

    case(EMOS2)
      numPolyPts = m2NumPolyPts
      allocate(poly1X(numPolyPts), poly1Y(numPolyPts))
      allocate(poly2X(numPolyPts), poly2Y(numPolyPts))
      poly1X = m2Poly1X
      poly2X = m2Poly2X
      poly1Y = m2Poly1Y
      poly2Y = m2Poly2Y

    case default
call error('',"")
    end select

    call readWcs(templateImageSetName, wcs) ! in ssclib/wcs_aux.

    call splitSetTabName(srcListSetTabName, srcListSetName, srcListTabName)
    ! in ssclib/dal_aux.
    srcListSet = clone(srcListSetName, tempSetName, TEMP)
    srcListTab = table(srcListSet, srcListTabName, MODIFY)
    call initFilter(selectionExpression)
    numSources = doFilter(srcListTab)
    if (numSources <= 0)&
call error('noSourcesSelected'&
, "No sources passed the specified filtering expression.")

    allocate(srcRa(  numSources), srcDec( numSources))
    allocate(srcDetX(numSources), srcDetY(numSources))
    allocate(detX(numPolyPts,numSources*2), detY(numPolyPts,numSources*2))
    allocate(tanX(numPolyPts,numSources*2), tanY(numPolyPts,numSources*2))
    allocate(pixX(numPolyPts,numSources*2), pixY(numPolyPts,numSources*2))

    srcRaDeg  => real64Data(column(srcListTab, 'RA'))
    srcDecDeg => real64Data(column(srcListTab, 'DEC'))
    srcRa  = srcRaDeg  * PiOn180
    srcDec = srcDecDeg * PiOn180
    call release(srcListSet)

    ! Coordinate transforms:

    call raDecToDet(srcRa, srcDec, scAttitude, expStartTimeStamp&
      , srcDetX, srcDetY) ! in ssclib/coordinate.

    row = 1
    do sourceNum = 1, numSources
      detX(:,row) = srcDetX(sourceNum) + poly1X
      detY(:,row) = srcDetY(sourceNum) + poly1Y
      row = row + 1
      detX(:,row) = srcDetX(sourceNum) + poly2X
      detY(:,row) = srcDetY(sourceNum) + poly2Y
      row = row + 1
    end do

    call detToTan(detX, detY, wcs%x%refWorld*PiOn180, wcs%y%refWorld*PiOn180&
      , scAttitude, expStartTimeStamp, tanX, tanY) ! in ssclib/coordinate.

    call tanToPixels(wcs, tanX, tanY, pixX, pixY) ! in ssclib/wcs_aux.

    ! Save the output regions:

!***** can we call region/ims_ascregion_mod.f90:writeRegionExtension instead??

    regionSet = dataSet(regionSetName, CREATE)
    regionTab = addtable(regionSet, 'REGION', numSources*2)

    call setAttribute(regionTab, "HDUVERS", "1.0.0")
    call setAttribute(regionTab, "HDUCLASS", "ASC")
    call setAttribute(regionTab, "HDUCLAS1", "REGION")
    call setAttribute(regionTab, "HDUCLAS2", "STANDARD")
    call setAttribute(regionTab, "MTYPE1", "pos")
    call setAttribute(regionTab, "MFORM1", "X,Y")

    shapeCol = addColumn(regionTab, 'SHAPE', STRING, dimensions=(/8/))
    xCol     = addColumn(regionTab, 'X',     REAL32, dimensions=(/numPolyPts/))
    yCol     = addColumn(regionTab, 'Y',     REAL32, dimensions=(/numPolyPts/))

!*** call writeWcs(xCol, yCol, wcs) ! in ssclib/wcs_aux.
!*** need this? check with regionmask

    do row = 0, numSources*2-1
      call setStringCell(shapeCol, row, shapeName)
    end do

    xData => real32VectorData(xCol)
    yData => real32VectorData(yCol)
    xData = pixX
    yData = pixY

    call release(regionSet)

  end subroutine estrayregion_future

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

eventSetName    = stringParameter('eventset')
    srcListSetTabName    = stringParameter('srclisttab')
    selectionExpression  = stringParameter('expression')
    templateImageSetName = stringParameter('templateset')
    regionSetName        = stringParameter('regionset')
    tempSetName          = stringParameter('tempset')
    radiusArcsec         = realParameter('cutrad')

    applyAttcalcFudge = booleanParameter('withboresightfudge')
    ! only needed while boresight psi has wrong sign. IMS 2005-11-17

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template

end module estrayregion_mod

